import textwrap

from ...database.utils import start_bit


class Colors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def signal_tree_string(message, console_width=80, with_comments=False):
    """Returns the message signal tree as a string.

    """

    def get_prefix(index, length):
        if index < length - 1:
            return '|   '
        else:
            return '    '

    def add_prefix(prefix, lines):
        return [prefix + line for line in lines]

    def format_signal_line(signal_name):
        siginst = message.get_signal_by_name(signal_name)
        signal_name_line = signal_name

        if with_comments:
            com = []
            if siginst.comment:
                com.append(siginst.comment)
            if siginst.unit:
                com.append(f'[{siginst.unit}]')

            comstr = ' '.join(com)
            if len(comstr) > 0:
                signal_name_line = f'{signal_name} {Colors.OKBLUE}{comstr}{Colors.ENDC}'

        signal_name_line = textwrap.wrap(signal_name_line, width=console_width - 2, initial_indent='+-- ',
                                         subsequent_indent=(' ' * (8 + len(signal_name))))
        signal_name_line = '\n'.join(signal_name_line)

        return signal_name_line

    def format_mux(mux):
        signal_name, multiplexed_signals = next(iter(mux.items()))
        selector_signal = message.get_signal_by_name(signal_name)
        multiplexed_signals = sorted(multiplexed_signals.items())
        lines = []

        for index, multiplexed_signal in enumerate(multiplexed_signals):
            multiplexer_id, signal_names = multiplexed_signal
            multiplexer_desc = f'{multiplexer_id}'

            if selector_signal.choices and \
               multiplexer_id in selector_signal.choices:
                multiplexer_desc = \
                    f'{selector_signal.choices[multiplexer_id]} ' \
                    f'({multiplexer_id})'

            lines.append(f'+-- {multiplexer_desc}')
            lines += add_prefix(get_prefix(index, len(multiplexed_signals)),
                                format_level_lines(signal_names))

        return format_signal_line(signal_name), lines

    def format_level_lines(signal_names):
        lines = []

        for index, signal_name in enumerate(signal_names):
            if isinstance(signal_name, dict):
                signal_name_line, signal_lines = format_mux(signal_name)
                signal_lines = add_prefix(get_prefix(index, len(signal_names)),
                                          signal_lines)
            else:
                signal_name_line = format_signal_line(signal_name)
                signal_lines = []

            lines.append(signal_name_line)
            lines += signal_lines

        return lines

    lines = format_level_lines(message.signal_tree)
    lines = ['-- {root}', *add_prefix('   ', lines)]

    return '\n'.join(lines)


def layout_string(message, signal_names=True):
    """Returns the message layout as an ASCII art string. Each signal is
    an arrow from LSB ``x`` to MSB ``<``. Overlapping signal bits
    are set to ``X``.

    Set `signal_names` to ``False`` to hide signal names.

    .. code:: text

                            Bit

                7   6   5   4   3   2   1   0
                +---+---+---+---+---+---+---+---+
            0 |   |   |   |   |   |<----------|
                +---+---+---+---+---+---+---+---+
            1 |------x|   |   |   |   |<-x|   |
                +---+---+---+---+---+---+---+---+
                    |                   +-- Bar
                    +-- Foo
                +---+---+---+---+---+---+---+---+
            2 |   |   |   |   |   |   |   |   |
        B     +---+---+---+---+---+---+---+---+
        y   3 |----XXXXXXX---x|   |   |   |   |
        t     +---+---+---+---+---+---+---+---+
        e           +-- Fie
                +---+---+---+---+---+---+---+---+
            4 |-------------------------------|
                +---+---+---+---+---+---+---+---+
            5 |   |   |<----------------------|
                +---+---+---+---+---+---+---+---+
                        +-- Fum
                +---+---+---+---+---+---+---+---+
            6 |   |   |   |   |   |   |   |   |
                +---+---+---+---+---+---+---+---+
            7 |   |   |   |   |   |   |   |   |
                +---+---+---+---+---+---+---+---+

    """

    def format_big():
        signals = []

        for signal in message._signals:
            if signal.byte_order != 'big_endian':
                continue

            formatted = start_bit(signal) * '   '
            formatted += '<{}x'.format((3 * signal.length - 2) * '-')
            signals.append(formatted)

        return signals

    def format_little():
        signals = []

        for signal in message._signals:
            if signal.byte_order != 'little_endian':
                continue

            formatted = signal.start * '   '
            formatted += 'x{}<'.format((3 * signal.length - 2) * '-')
            end = signal.start + signal.length

            if end % 8 != 0:
                formatted += (8 - (end % 8)) * '   '

            formatted = ''.join([
                formatted[i:i + 24][::-1]
                for i in range(0, len(formatted), 24)
            ])
            signals.append(formatted)

        return signals

    def format_byte_lines():
        # Signal lines.
        signals = format_big() + format_little()

        if len(signals) > 0:
            length = max([len(signal) for signal in signals])

            if length % 24 != 0:
                length += (24 - (length % 24))

            signals = [signal + (length - len(signal)) * ' ' for signal in signals]

        # Signals union line.
        signals_union = ''

        for chars in zip(*signals, strict=False):
            head = chars.count('<')
            dash = chars.count('-')
            tail = chars.count('x')

            if head + dash + tail > 1:
                signals_union += 'X'
            elif head == 1:
                signals_union += '<'
            elif dash == 1:
                signals_union += '-'
            elif tail == 1:
                signals_union += 'x'
            else:
                signals_union += ' '

        # Split the signals union line into byte lines, 8 bits per
        # line.
        byte_lines = [
            signals_union[i:i + 24]
            for i in range(0, len(signals_union), 24)
        ]

        unused_byte_lines = (message._length - len(byte_lines))

        if unused_byte_lines > 0:
            byte_lines += unused_byte_lines * [24 * ' ']

        # Insert bits separators into each byte line.
        lines = []

        for byte_line in byte_lines:
            line = ''
            prev_byte = None

            for i in range(0, 24, 3):
                byte_triple = byte_line[i:i + 3]

                if i == 0 or byte_triple[0] in ' <>x':
                    line += '|'
                elif byte_triple[0] == 'X':
                    if prev_byte == 'X':
                        line += 'X'
                    elif prev_byte == '-':
                        line += '-'
                    else:
                        line += '|'
                else:
                    line += '-'

                line += byte_triple
                prev_byte = byte_triple[2]

            line += '|'
            lines.append(line)

        # Add byte numbering.
        number_width = len(str(len(lines))) + 4
        number_fmt = f'{{:{number_width - 1}d}} {{}}'
        a = []

        for number, line in enumerate(lines):
            a.append(number_fmt.format(number, line))

        return a, len(lines), number_width

    def add_header_lines(lines, number_width):
        padding = number_width * ' '

        return [
            padding + '               Bit',
            padding + '',
            padding + '  7   6   5   4   3   2   1   0',
            padding + '+---+---+---+---+---+---+---+---+',
            *lines,
        ]

    def add_horizontal_lines(byte_lines, number_width):
        padding = number_width * ' '
        lines = []

        for byte_line in byte_lines:
            lines.append(byte_line)
            lines.append(padding + '+---+---+---+---+---+---+---+---+')

        return lines

    def name_bit(signal):
        offset = start_bit(signal) + signal.length - 1

        if signal.byte_order == 'big_endian':
            return (8 * (offset // 8) + (7 - (offset % 8)))
        else:
            return offset

    def add_signal_names(input_lines,
                         number_of_bytes,
                         number_width):
        # Find MSB and name of all signals.
        padding = number_width * ' '
        signals_per_byte = [[] for _ in range(number_of_bytes)]

        for signal in message._signals:
            byte, bit = divmod(name_bit(signal), 8)
            signals_per_byte[byte].append((bit, '+-- ' + signal.name))

        # Format signal lines.
        signal_lines_per_byte = []

        for signals in signals_per_byte:
            signals = sorted(signals)
            signals_lines = []

            for signal in signals:
                line = number_width * ' ' + '  ' + signal[1]
                line = (7 - signal[0]) * '    ' + line
                chars = list(line)

                for other_signal in signals:
                    if other_signal[0] > signal[0]:
                        other_signal_msb = (number_width
                                            + 2
                                            + 4 * (7 - other_signal[0]))
                        chars[other_signal_msb] = '|'

                signals_lines.append(''.join(chars))

            signal_lines_per_byte.append(signals_lines)

        # Insert the signals names lines among other lines.
        lines = []

        for number in range(number_of_bytes):
            lines += input_lines[2 * number: 2 * number + 2]

            if signal_lines_per_byte[number]:
                lines += signal_lines_per_byte[number]

                if number + 1 < number_of_bytes:
                    lines.append(
                        padding + '+---+---+---+---+---+---+---+---+')

        return lines

    def add_y_axis_name(lines):
        number_of_matrix_lines = (len(lines) - 3)

        if number_of_matrix_lines < 5:
            lines += (5 - number_of_matrix_lines) * ['     ']

        start_index = 4 + ((number_of_matrix_lines - 4) // 2 - 1)

        start_index = max(start_index, 4)

        axis_lines = start_index * ['  ']
        axis_lines += [' B', ' y', ' t', ' e']
        axis_lines += (len(lines) - start_index - 4) * ['  ']

        return [
            axis_line + line
            for axis_line, line in zip(axis_lines, lines, strict=False)
        ]

    lines, number_of_bytes, number_width = format_byte_lines()
    lines = add_horizontal_lines(lines, number_width)

    if signal_names:
        lines = add_signal_names(lines,
                                 number_of_bytes,
                                 number_width)

    lines = add_header_lines(lines, number_width)
    lines = add_y_axis_name(lines)
    lines = [line.rstrip() for line in lines]

    return '\n'.join(lines)


def signal_choices_string(message):
    """Returns the signal choices as a string.

    """

    lines = []

    for signal in message._signals:
        if signal.choices:
            lines.append('')
            lines.append(signal.name)

            for value, text in sorted(signal.choices.items()):
                lines.append(f'    {value} {text}')

    return '\n'.join(lines)





class AutosarEnd2EndProperties:
    """This class collects all attributes that are required to implement
    AUTOSAR-conformant End-to-End protection (CRCs) of messages
    """

    def __init__(self) -> None:
        self._category: str | None = None
        self._data_ids: list[int] | None = None
        self._payload_length: int = 0

    @property
    def category(self) -> str | None:
        """The category string of the applicable end-to-end protection
        mechanism

        Note that the contents of these are not specified by the
        AUTOSAR standard.
        """
        return self._category

    @category.setter
    def category(self, value: str | None) -> None:
        self._category = value

    @property
    def data_ids(self) -> list[int] | None:
        """The list of data IDs applicable
        """
        return self._data_ids

    @data_ids.setter
    def data_ids(self, value: list[int] | None) -> None:
        self._data_ids = value

    @property
    def payload_length(self) -> int:
        """The size of the end-to-end protected data in bytes

        This number includes the end-to-end protection signals
        themselves (i.e. the sequence counter and the CRC value)

        """
        return self._payload_length

    @payload_length.setter
    def payload_length(self, value: int) -> None:
        self._payload_length = value



# Load an ECU extract CAN database from an ARXML formatted file.
import logging
from typing import TYPE_CHECKING, Any

from ....conversion import BaseConversion
from ....utils import sort_signals_by_start_bit, type_sort_signals
from ...internal_database import InternalDatabase
from ...message import Message
from ...signal import Signal

if TYPE_CHECKING:
    from ...bus import Bus


def make_xpath(location: list[str]) -> str:
    """Convenience function to traverse the XML element tree more easily

    (This function is only used by the EcuExtractLoader.)"""
    return './ns:' + '/ns:'.join(location)

LOGGER = logging.getLogger(__name__)

# The ARXML XML namespace for the EcuExtractLoader
NAMESPACE = 'http://autosar.org/schema/r4.0'
NAMESPACES = {'ns': NAMESPACE}

ECUC_VALUE_COLLECTION_XPATH = make_xpath([
    'AR-PACKAGES',
    'AR-PACKAGE',
    'ELEMENTS',
    'ECUC-VALUE-COLLECTION'
])
ECUC_MODULE_CONFIGURATION_VALUES_REF_XPATH = make_xpath([
    'ECUC-VALUES',
    'ECUC-MODULE-CONFIGURATION-VALUES-REF-CONDITIONAL',
    'ECUC-MODULE-CONFIGURATION-VALUES-REF'
])
ECUC_REFERENCE_VALUE_XPATH = make_xpath([
    'REFERENCE-VALUES',
    'ECUC-REFERENCE-VALUE'
])
DEFINITION_REF_XPATH = make_xpath(['DEFINITION-REF'])
VALUE_XPATH = make_xpath(['VALUE'])
VALUE_REF_XPATH = make_xpath(['VALUE-REF'])
SHORT_NAME_XPATH = make_xpath(['SHORT-NAME'])
PARAMETER_VALUES_XPATH = make_xpath(['PARAMETER-VALUES'])
REFERENCE_VALUES_XPATH = make_xpath([
    'REFERENCE-VALUES'
])

class EcuExtractLoader:

    def __init__(self,
                 root:Any,
                 strict:bool,
                 sort_signals:type_sort_signals=sort_signals_by_start_bit):
        self.root = root
        self.strict = strict
        self.sort_signals = sort_signals

    def load(self) -> InternalDatabase:
        buses: list[Bus] = []
        messages = []
        version = None

        ecuc_value_collection = self.root.find(ECUC_VALUE_COLLECTION_XPATH,
                                               NAMESPACES)
        values_refs = ecuc_value_collection.iterfind(
            ECUC_MODULE_CONFIGURATION_VALUES_REF_XPATH,
            NAMESPACES)
        com_xpaths = [
            value_ref.text
            for value_ref in values_refs
            if value_ref.text.endswith('/Com')
        ]

        if len(com_xpaths) != 1:
            raise ValueError(
                f'Expected 1 /Com, but got {len(com_xpaths)}.')

        com_config = self.find_com_config(com_xpaths[0] + '/ComConfig')

        for ecuc_container_value in com_config:
            definition_ref = ecuc_container_value.find(DEFINITION_REF_XPATH,
                                                       NAMESPACES).text

            if not definition_ref.endswith('ComIPdu'):
                continue

            message = self.load_message(ecuc_container_value)

            if message is not None:
                messages.append(message)

        return InternalDatabase(messages,
                                [],
                                buses,
                                version)

    def load_message(self, com_i_pdu):
        # Default values.
        interval = None
        senders = []
        comments = None

        # Name, frame id, length and is_extended_frame.
        name = com_i_pdu.find(SHORT_NAME_XPATH, NAMESPACES).text
        direction = None

        for parameter, value in self.iter_parameter_values(com_i_pdu):
            if parameter == 'ComIPduDirection':
                direction = value
                break

        com_pdu_id_ref = None

        for reference, value in self.iter_reference_values(com_i_pdu):
            if reference == 'ComPduIdRef':
                com_pdu_id_ref = value
                break

        if com_pdu_id_ref is None:
            raise ValueError('No ComPduIdRef reference found.')

        if direction == 'SEND':
            frame_id, length, is_extended_frame = self.load_message_tx(
                com_pdu_id_ref)
        elif direction == 'RECEIVE':
            frame_id, length, is_extended_frame = self.load_message_rx(
                com_pdu_id_ref)
        else:
            raise NotImplementedError(
                f'Direction {direction} not supported.')

        if frame_id is None:
            LOGGER.warning('No frame id found for message %s.', name)

            return None

        if is_extended_frame is None:
            LOGGER.warning('No frame type found for message %s.', name)

            return None

        if length is None:
            LOGGER.warning('No length found for message %s.', name)

            return None

        # ToDo: interval, senders, comments

        # Find all signals in this message.
        signals = []
        values = com_i_pdu.iterfind(ECUC_REFERENCE_VALUE_XPATH,
                                    NAMESPACES)

        for value in values:
            definition_ref = value.find(DEFINITION_REF_XPATH,
                                        NAMESPACES).text
            if not definition_ref.endswith('ComIPduSignalRef'):
                continue

            value_ref = value.find(VALUE_REF_XPATH, NAMESPACES)
            signal = self.load_signal(value_ref.text)

            if signal is not None:
                signals.append(signal)

        return Message(frame_id=frame_id,
                       is_extended_frame=is_extended_frame,
                       name=name,
                       length=length,
                       senders=senders,
                       send_type=None,
                       cycle_time=interval,
                       signals=signals,
                       comment=comments,
                       bus_name=None,
                       strict=self.strict,
                       sort_signals=self.sort_signals)

    def load_message_tx(self, com_pdu_id_ref):
        return self.load_message_rx_tx(com_pdu_id_ref,
                                       'CanIfTxPduCanId',
                                       'CanIfTxPduDlc',
                                       'CanIfTxPduCanIdType')

    def load_message_rx(self, com_pdu_id_ref):
        return self.load_message_rx_tx(com_pdu_id_ref,
                                       'CanIfRxPduCanId',
                                       'CanIfRxPduDlc',
                                       'CanIfRxPduCanIdType')

    def load_message_rx_tx(self,
                           com_pdu_id_ref,
                           parameter_can_id,
                           parameter_dlc,
                           parameter_can_id_type):
        can_if_tx_pdu_cfg = self.find_can_if_rx_tx_pdu_cfg(com_pdu_id_ref)
        frame_id = None
        length = None
        is_extended_frame = None

        if can_if_tx_pdu_cfg is not None:
            for parameter, value in self.iter_parameter_values(can_if_tx_pdu_cfg):
                if parameter == parameter_can_id:
                    frame_id = int(value)
                elif parameter == parameter_dlc:
                    length = int(value)
                elif parameter == parameter_can_id_type:
                    is_extended_frame = (value == 'EXTENDED_CAN')

        return frame_id, length, is_extended_frame

    def load_signal(self, xpath):
        ecuc_container_value = self.find_value(xpath)
        if ecuc_container_value is None:
            return None

        name = ecuc_container_value.find(SHORT_NAME_XPATH, NAMESPACES).text

        # Default values.
        is_signed = False
        is_float = False
        minimum = None
        maximum = None
        factor = 1.0
        offset = 0.0
        unit = None
        choices = None
        comments = None
        receivers = []

        # Bit position, length, byte order, is_signed and is_float.
        bit_position = None
        length = None
        byte_order = None

        for parameter, value in self.iter_parameter_values(ecuc_container_value):
            if parameter == 'ComBitPosition':
                bit_position = int(value)
            elif parameter == 'ComBitSize':
                length = int(value)
            elif parameter == 'ComSignalEndianness':
                byte_order = value.lower()
            elif parameter == 'ComSignalType':
                if value in ['SINT8', 'SINT16', 'SINT32']:
                    is_signed = True
                elif value in ['FLOAT32', 'FLOAT64']:
                    is_float = True

        if bit_position is None:
            LOGGER.warning('No bit position found for signal %s.',name)

            return None

        if length is None:
            LOGGER.warning('No bit size found for signal %s.', name)

            return None

        if byte_order is None:
            LOGGER.warning('No endianness found for signal %s.', name)

            return None

        # ToDo: minimum, maximum, factor, offset, unit, choices,
        #       comments and receivers.

        conversion = BaseConversion.factory(
            scale=factor,
            offset=offset,
            choices=choices,
            is_float=is_float,
        )

        return Signal(name=name,
                      start=bit_position,
                      length=length,
                      receivers=receivers,
                      byte_order=byte_order,
                      is_signed=is_signed,
                      conversion=conversion,
                      minimum=minimum,
                      maximum=maximum,
                      unit=unit,
                      comment=comments,
                      )

    def find_com_config(self, xpath):
        return self.root.find(make_xpath([
            "AR-PACKAGES",
            "AR-PACKAGE/[ns:SHORT-NAME='{}']".format(xpath.split('/')[1]),
            "ELEMENTS",
            "ECUC-MODULE-CONFIGURATION-VALUES/[ns:SHORT-NAME='Com']",
            "CONTAINERS",
            "ECUC-CONTAINER-VALUE/[ns:SHORT-NAME='ComConfig']",
            "SUB-CONTAINERS"
        ]),
                              NAMESPACES)

    def find_value(self, xpath):
        return self.root.find(make_xpath([
            "AR-PACKAGES",
            "AR-PACKAGE/[ns:SHORT-NAME='{}']".format(xpath.split('/')[1]),
            "ELEMENTS",
            "ECUC-MODULE-CONFIGURATION-VALUES/[ns:SHORT-NAME='Com']",
            "CONTAINERS",
            "ECUC-CONTAINER-VALUE/[ns:SHORT-NAME='ComConfig']",
            "SUB-CONTAINERS",
            "ECUC-CONTAINER-VALUE/[ns:SHORT-NAME='{}']".format(xpath.split('/')[-1])
        ]),
                              NAMESPACES)

    def find_can_if_rx_tx_pdu_cfg(self, com_pdu_id_ref):
        messages = self.root.iterfind(
            make_xpath([
                "AR-PACKAGES",
                "AR-PACKAGE/[ns:SHORT-NAME='{}']".format(
                    com_pdu_id_ref.split('/')[1]),
                "ELEMENTS",
                "ECUC-MODULE-CONFIGURATION-VALUES/[ns:SHORT-NAME='CanIf']",
                'CONTAINERS',
                "ECUC-CONTAINER-VALUE/[ns:SHORT-NAME='CanIfInitCfg']",
                'SUB-CONTAINERS',
                'ECUC-CONTAINER-VALUE'
            ]),
            NAMESPACES)

        for message in messages:
            definition_ref = message.find(DEFINITION_REF_XPATH,
                                          NAMESPACES).text

            if definition_ref.endswith('CanIfTxPduCfg'):
                expected_reference = 'CanIfTxPduRef'
            elif definition_ref.endswith('CanIfRxPduCfg'):
                expected_reference = 'CanIfRxPduRef'
            else:
                continue

            for reference, value in self.iter_reference_values(message):
                if reference == expected_reference and value == com_pdu_id_ref:
                    return message

    def iter_parameter_values(self, param_conf_container):
        parameters = param_conf_container.find(PARAMETER_VALUES_XPATH,
                                               NAMESPACES)

        if parameters is None:
            raise ValueError('PARAMETER-VALUES does not exist.')

        for parameter in parameters:
            definition_ref = parameter.find(DEFINITION_REF_XPATH,
                                            NAMESPACES).text
            value = parameter.find(VALUE_XPATH, NAMESPACES).text
            name = definition_ref.split('/')[-1]

            yield name, value

    def iter_reference_values(self, param_conf_container):
        references = param_conf_container.find(REFERENCE_VALUES_XPATH,
                                               NAMESPACES)

        if references is None:
            raise ValueError('REFERENCE-VALUES does not exist.')

        for reference in references:
            definition_ref = reference.find(DEFINITION_REF_XPATH,
                                            NAMESPACES).text
            value = reference.find(VALUE_REF_XPATH, NAMESPACES).text
            name = definition_ref.split('/')[-1]

            yield name, value



import re
import time
import warnings
from collections.abc import Iterator
from typing import (
    TYPE_CHECKING,
    TypeVar,
    cast,
)

from cantools import __version__
from cantools.database.can.message import SignalTreeMuxElemType, SignalTreeType

if TYPE_CHECKING:
    from cantools.database.can import Database, Message, Signal


_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")
THelperKind = tuple[str, int]


HEADER_FMT = '''\
/**
 * @file {file_name}
 *
 * @brief This header file was generated by cantools version {version} {date}.
 *
 * @copyright Copyright (c) 2018-2019 Erik Moqvist
 *
 * @par License
 * The MIT License (MIT)
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef {include_guard}
#define {include_guard}

#ifdef __cplusplus
extern "C" {{
#endif

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#ifndef EINVAL
#    define EINVAL 22
#endif

/* Frame ids. */
{frame_id_defines}

/* Frame lengths in bytes. */
{frame_length_defines}

/* Extended or standard frame types. */
{is_extended_frame_defines}

/* Frame cycle times in milliseconds. */
{frame_cycle_time_defines}

/* Signal choices. */
{choices_defines}

/* Frame Names. */
{frame_name_macros}

/* Signal Names. */
{signal_name_macros}

{structs}
{declarations}

#ifdef __cplusplus
}}
#endif

#endif
'''

SOURCE_FMT = '''\
/**
 * @file {file_name}
 *
 * @brief This source file was generated by cantools version {version} {date}.
 *
 * @copyright Copyright (c) 2018-2019 Erik Moqvist
 *
 * @par License
 * The MIT License (MIT)
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

{includes}\
#include <string.h>

#include "{header}"

{helpers}\
{definitions}\
'''

FUZZER_SOURCE_FMT = '''\
/**
 * The MIT License (MIT)
 *
 * Copyright (c) 2018-2019 Erik Moqvist
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/**
 * This file was generated by cantools version {version} {date}.
 */

#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>

#include "{header}"

static void assert_first_pack(int res)
{{
    if (res < 0) {{
        printf("First pack failed with %ld.\\n", res);
        __builtin_trap();
    }}
}}

static void assert_second_unpack(int res)
{{
    if (res < 0) {{
        printf("Second unpack failed with %ld.\\n", res);
        __builtin_trap();
    }}
}}

static void assert_second_unpack_data(const void *unpacked_p,
                                      const void *unpacked2_p,
                                      size_t size)
{{
    if (memcmp(unpacked_p, unpacked2_p, size) != 0) {{
        printf("Second unpacked data does not match first unpacked data.\\n");
        __builtin_trap();
    }}
}}

static void assert_second_pack(int res, int res2)
{{
    if (res != res2) {{
        printf("Second pack result %ld does not match first pack "
               "result %ld.\\n",
               res,
               res2);
        __builtin_trap();
    }}
}}

static void assert_second_pack_data(const uint8_t *packed_p,
                                    const uint8_t *packed2_p,
                                    int size)
{{
    int i;

    if (memcmp(packed_p, packed2_p, size) != 0) {{
        for (i = 0; i < size; i++) {{
            printf("[%04ld]: 0x%02x 0x%02x\\n", i, packed_p[i], packed2_p[i]);
        }}

        __builtin_trap();
    }}
}}

{tests}

int LLVMFuzzerTestOneInput(const uint8_t *data_p, size_t size)
{{
{llvm_body}

    return (0);
}}
'''

FUZZER_MAKEFILE_FMT = '''\
#
# The MIT License (MIT)
#
# Copyright (c) 2018-2019 Erik Moqvist
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use, copy,
# modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

#
# This file was generated by cantools version {version} {date}.
#

CC = clang
EXE = fuzzer
C_SOURCES = \\
\t{source} \\
\t{fuzzer_source}
CFLAGS = \\
\t-fprofile-instr-generate \\
\t-fcoverage-mapping \\
\t-I. \\
\t-g -fsanitize=address,fuzzer \\
\t-fsanitize=signed-integer-overflow \\
\t-fno-sanitize-recover=all
EXECUTION_TIME ?= 5

all:
\t$(CC) $(CFLAGS) $(C_SOURCES) -o $(EXE)
\trm -f $(EXE).profraw
\tLLVM_PROFILE_FILE="$(EXE).profraw" \\
\t    ./$(EXE) \\
\t    -max_total_time=$(EXECUTION_TIME)
\tllvm-profdata merge -sparse $(EXE).profraw -o $(EXE).profdata
\tllvm-cov show ./$(EXE) -instr-profile=$(EXE).profdata
\tllvm-cov report ./$(EXE) -instr-profile=$(EXE).profdata

'''

TEST_FMT = '''
static void test_{name}(
    const uint8_t *packed_p,
    size_t size)
{{
    int res;
    int res2;
    uint8_t packed[size];
    uint8_t packed2[size];
    struct {name}_t unpacked;
    struct {name}_t unpacked2;

    memset(&unpacked, 0, sizeof(unpacked));

    res = {name}_unpack(
        &unpacked,
        packed_p,
        size);

    if (res >= 0) {{
        res = {name}_pack(
            &packed[0],
            &unpacked,
            sizeof(packed));

        assert_first_pack(res);

        memset(&unpacked2, 0, sizeof(unpacked2));

        res2 = {name}_unpack(
            &unpacked2,
            &packed[0],
            res);

        assert_second_unpack(res2);
        assert_second_unpack_data(&unpacked,
                                  &unpacked2,
                                  sizeof(unpacked));

        res2 = {name}_pack(
            &packed2[0],
            &unpacked,
            sizeof(packed2));

        assert_second_pack(res, res2);
        assert_second_pack_data(&packed[0], &packed2[0], res);
    }}
}}\
'''

STRUCT_FMT = '''\
/**
 * Signals in message {database_message_name}.
 *
{comment}\
 * All signal values are as on the CAN bus.
 */
struct {database_name}_{message_name}_t {{
{members}
}};
'''

DECLARATION_PACK_FMT = '''\
/**
 * Pack message {database_message_name}.
 *
 * @param[out] dst_p Buffer to pack the message into.
 * @param[in] src_p Data to pack.
 * @param[in] size Size of dst_p.
 *
 * @return Size of packed data, or negative error code.
 */
int {database_name}_{message_name}_pack(
    uint8_t *dst_p,
    const struct {database_name}_{message_name}_t *src_p,
    size_t size);

'''

DECLARATION_UNPACK_FMT = '''\
/**
 * Unpack message {database_message_name}.
 *
 * @param[out] dst_p Object to unpack the message into.
 * @param[in] src_p Message to unpack.
 * @param[in] size Size of src_p.
 *
 * @return zero(0) or negative error code.
 */
int {database_name}_{message_name}_unpack(
    struct {database_name}_{message_name}_t *dst_p,
    const uint8_t *src_p,
    size_t size);

'''

SIGNAL_DECLARATION_ENCODE_FMT = '''\
/**
 * Encode given signal by applying scaling and offset.
 *
 * @param[in] value Signal to encode.
 *
 * @return Encoded signal.
 */
{type_name} {database_name}_{message_name}_{signal_name}_encode({floating_point_type} value);

'''

SIGNAL_DECLARATION_DECODE_FMT = '''\
/**
 * Decode given signal by applying scaling and offset.
 *
 * @param[in] value Signal to decode.
 *
 * @return Decoded signal.
 */
{floating_point_type} {database_name}_{message_name}_{signal_name}_decode({type_name} value);

'''

SIGNAL_DECLARATION_IS_IN_RANGE_FMT = '''\
/**
 * Check that given signal is in allowed range.
 *
 * @param[in] value Signal to check.
 *
 * @return true if in range, false otherwise.
 */
bool {database_name}_{message_name}_{signal_name}_is_in_range({type_name} value);
'''

SIGNAL_DECLARATION_IS_IN_PHYS_RANGE_FMT = '''\

/**
 * Check that given physical value is in allowed range before encoding.
 *
 * Use this BEFORE calling _encode() to avoid silent integer overflow:
 *   encode(-1.0) on a [0,100] signal wraps to 255 and passes is_in_range().
 *   is_in_phys_range(-1.0) correctly returns false.
 *
 * @param[in] value Physical signal value to check.
 *
 * @return true if in range, false otherwise.
 */
bool {database_name}_{message_name}_{signal_name}_is_in_phys_range({floating_point_type} value);
'''

MESSAGE_DECLARATION_INIT_FMT = '''\
/**
 * Init message fields to default values from {database_message_name}.
 *
 * @param[in] msg_p Message to init.
 *
 * @return zero(0) on success or (-1) in case of nullptr argument.
 */
int {database_name}_{message_name}_init(struct {database_name}_{message_name}_t *msg_p);
'''

MESSAGE_DEFINITION_INIT_FMT = '''\
int {database_name}_{message_name}_init(struct {database_name}_{message_name}_t *msg_p)
{{
    if (msg_p == NULL) return -1;

    memset(msg_p, 0, sizeof(struct {database_name}_{message_name}_t));
{init_body}
    return 0;
}}
'''

PACK_HELPER_LEFT_SHIFT_FMT = '''\
static inline uint8_t pack_left_shift_u{length}(
    {var_type} value,
    uint8_t shift,
    uint8_t mask)
{{
    return (uint8_t)((uint8_t)(value << shift) & mask);
}}
'''

PACK_HELPER_RIGHT_SHIFT_FMT = '''\
static inline uint8_t pack_right_shift_u{length}(
    {var_type} value,
    uint8_t shift,
    uint8_t mask)
{{
    return (uint8_t)((uint8_t)(value >> shift) & mask);
}}
'''

UNPACK_HELPER_LEFT_SHIFT_FMT = '''\
static inline {var_type} unpack_left_shift_u{length}(
    uint8_t value,
    uint8_t shift,
    uint8_t mask)
{{
    return ({var_type})(({var_type})(value & mask) << shift);
}}
'''

UNPACK_HELPER_RIGHT_SHIFT_FMT = '''\
static inline {var_type} unpack_right_shift_u{length}(
    uint8_t value,
    uint8_t shift,
    uint8_t mask)
{{
    return ({var_type})(({var_type})(value & mask) >> shift);
}}
'''

DEFINITION_PACK_FMT = '''\
int {database_name}_{message_name}_pack(
    uint8_t *dst_p,
    const struct {database_name}_{message_name}_t *src_p,
    size_t size)
{{
{pack_unused}\
{pack_variables}\
    if (size < {message_length}u) {{
        return (-EINVAL);
    }}

    memset(&dst_p[0], 0, {message_length});
{pack_body}
    return ({message_length});
}}

'''

DEFINITION_UNPACK_FMT = '''\
int {database_name}_{message_name}_unpack(
    struct {database_name}_{message_name}_t *dst_p,
    const uint8_t *src_p,
    size_t size)
{{
{unpack_unused}\
{unpack_variables}\
    if (size < {message_length}u) {{
        return (-EINVAL);
    }}
{unpack_body}
    return (0);
}}

'''

SIGNAL_DEFINITION_ENCODE_FMT = '''\
{type_name} {database_name}_{message_name}_{signal_name}_encode({floating_point_type} value)
{{
    return ({type_name})({encode});
}}

'''

SIGNAL_DEFINITION_DECODE_FMT = '''\
{floating_point_type} {database_name}_{message_name}_{signal_name}_decode({type_name} value)
{{
    return ({decode});
}}

'''

SIGNAL_DEFINITION_IS_IN_RANGE_FMT = '''\
bool {database_name}_{message_name}_{signal_name}_is_in_range({type_name} value)
{{
{unused}\
    return ({check});
}}
'''

SIGNAL_DEFINITION_IS_IN_PHYS_RANGE_FMT = '''\

bool {database_name}_{message_name}_{signal_name}_is_in_phys_range({floating_point_type} value)
{{
{phys_unused}\
    return ({phys_check});
}}
'''

EMPTY_DEFINITION_FMT = '''\
int {database_name}_{message_name}_pack(
    uint8_t *dst_p,
    const struct {database_name}_{message_name}_t *src_p,
    size_t size)
{{
    (void)dst_p;
    (void)src_p;
    (void)size;

    return (0);
}}

int {database_name}_{message_name}_unpack(
    struct {database_name}_{message_name}_t *dst_p,
    const uint8_t *src_p,
    size_t size)
{{
    (void)dst_p;
    (void)src_p;
    (void)size;

    return (0);
}}
'''

SIGN_EXTENSION_FMT = '''
    if (({name} & (1{suffix} << {shift})) != 0{suffix}) {{
        {name} |= 0x{mask:x}{suffix};
    }}

'''

SIGNAL_MEMBER_FMT = '''\
    /**
{comment}\
     * Range: {range}
     * Scale: {scale}
     * Offset: {offset}
     */
    {type_name} {name}{length};\
'''

INIT_SIGNAL_BODY_TEMPLATE_FMT = '''\
    msg_p->{signal_name} = {signal_initial};
'''


class CodeGenSignal:

    def __init__(self, signal: "Signal") -> None:
        self.signal: Signal = signal
        self.snake_name = camel_to_snake_case(signal.name)

    @property
    def unit(self) -> str:
        return _get(self.signal.unit, '-')

    @property
    def type_length(self) -> int:
        if self.signal.length <= 8:
            return 8
        elif self.signal.length <= 16:
            return 16
        elif self.signal.length <= 32:
            return 32
        else:
            return 64

    @property
    def type_name(self) -> str:
        if self.signal.conversion.is_float:
            if self.signal.length == 32:
                type_name = 'float'
            else:
                type_name = 'double'
        else:
            type_name = f'int{self.type_length}_t'

            if not self.signal.is_signed:
                type_name = 'u' + type_name

        return type_name

    @property
    def type_suffix(self) -> str:
        try:
            return {
                'uint8_t': 'u',
                'uint16_t': 'u',
                'uint32_t': 'u',
                'int64_t': 'll',
                'uint64_t': 'ull',
                'float': 'f'
            }[self.type_name]
        except KeyError:
            return ''

    @property
    def conversion_type_suffix(self) -> str:
        try:
            return {
                8: 'u',
                16: 'u',
                32: 'u',
                64: 'ull'
            }[self.type_length]
        except KeyError:
            return ''

    @property
    def unique_choices(self) -> dict[int, str]:
        """Make duplicated choice names unique by first appending its value
        and then underscores until unique.

        """
        if self.signal.choices is None:
            return {}

        items = {
            value: camel_to_snake_case(str(name)).upper()
            for value, name in self.signal.choices.items()
        }
        names = list(items.values())
        duplicated_names = [
            name
            for name in set(names)
            if names.count(name) > 1
        ]
        unique_choices = {
            value: name
            for value, name in items.items()
            if names.count(name) == 1
        }

        for value, name in items.items():
            if name in duplicated_names:
                name += _canonical(f'_{value}')

                while name in unique_choices.values():
                    name += '_'

                unique_choices[value] = name

        return unique_choices

    @property
    def minimum_ctype_value(self) -> int | None:
        if self.type_name == 'int8_t':
            return -2**7
        elif self.type_name == 'int16_t':
            return -2**15
        elif self.type_name == 'int32_t':
            return -2**31
        elif self.type_name == 'int64_t':
            return -2**63
        elif self.type_name.startswith('u'):
            return 0
        else:
            return None

    @property
    def maximum_ctype_value(self) -> int | None:
        if self.type_name == 'int8_t':
            return 2**7 - 1
        elif self.type_name == 'int16_t':
            return 2**15 - 1
        elif self.type_name == 'int32_t':
            return 2**31 - 1
        elif self.type_name == 'int64_t':
            return 2**63 - 1
        elif self.type_name == 'uint8_t':
            return 2**8 - 1
        elif self.type_name == 'uint16_t':
            return 2**16 - 1
        elif self.type_name == 'uint32_t':
            return 2**32 - 1
        elif self.type_name == 'uint64_t':
            return 2**64 - 1
        else:
            return None

    @property
    def minimum_can_raw_value(self) -> int | None:
        if self.signal.conversion.is_float:
            return None
        elif self.signal.is_signed:
            return cast("int", -(2 ** (self.signal.length - 1)))
        else:
            return 0

    @property
    def maximum_can_raw_value(self) -> int | None:
        if self.signal.conversion.is_float:
            return None
        elif self.signal.is_signed:
            return cast("int", (2 ** (self.signal.length - 1)) - 1)
        else:
            return cast("int", (2 ** self.signal.length) - 1)

    def segments(self, invert_shift: bool) -> Iterator[tuple[int, int, str, int]]:
        index, pos = divmod(self.signal.start, 8)
        left = self.signal.length

        while left > 0:
            if self.signal.byte_order == 'big_endian':
                if left >= (pos + 1):
                    length = (pos + 1)
                    pos = 7
                    shift = -(left - length)
                    mask = ((1 << length) - 1)
                else:
                    length = left
                    shift = (pos - length + 1)
                    mask = ((1 << length) - 1)
                    mask <<= (pos - length + 1)
            else:
                shift = (left - self.signal.length) + pos

                if left >= (8 - pos):
                    length = (8 - pos)
                    mask = ((1 << length) - 1)
                    mask <<= pos
                    pos = 0
                else:
                    length = left
                    mask = ((1 << length) - 1)
                    mask <<= pos

            if invert_shift:
                if shift < 0:
                    shift = -shift
                    shift_direction = 'left'
                else:
                    shift_direction = 'right'
            elif shift < 0:
                shift = -shift
                shift_direction = 'right'
            else:
                shift_direction = 'left'

            yield index, shift, shift_direction, mask

            left -= length
            index += 1


class CodeGenMessage:

    def __init__(self, message: "Message") -> None:
        self.message = message
        self.snake_name = camel_to_snake_case(message.name)
        self.cg_signals = [CodeGenSignal(signal) for signal in message.signals]

    def get_signal_by_name(self, name: str) -> "CodeGenSignal":
        for cg_signal in self.cg_signals:
            if cg_signal.signal.name == name:
                return cg_signal
        raise KeyError(f"Signal {name} not found.")


def _canonical(value: str) -> str:
    """Replace anything but 'a-z', 'A-Z' and '0-9' with '_'.

    """

    return re.sub(r'[^a-zA-Z0-9]', '_', value)


def camel_to_snake_case(value: str) -> str:
    value = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', value)
    value = re.sub(r'(_+)', '_', value)
    value = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', value).lower()
    value = _canonical(value)

    return value


def _strip_blank_lines(lines: list[str]) -> list[str]:
    try:
        while lines[0] == '':
            lines = lines[1:]

        while lines[-1] == '':
            lines = lines[:-1]
    except IndexError:
        pass

    return lines


def _get(value: _T1 | None, default: _T2) -> _T1 | _T2:
    if value is None:
        return default
    return value


def _format_comment(comment: str | None) -> str:
    if comment:
        return '\n'.join([
            '     * ' + line.rstrip()
            for line in comment.splitlines()
        ]) + '\n     *\n'
    else:
        return ''


def _format_range(cg_signal: "CodeGenSignal") -> str:
    minimum = cg_signal.signal.minimum
    maximum = cg_signal.signal.maximum

    def phys_to_raw(x: float) -> int | float:
        raw_val = cg_signal.signal.scaled_to_raw(x)
        if cg_signal.signal.is_float:
            return float(raw_val)
        return round(raw_val)

    if minimum is not None and maximum is not None:
        return \
            f'{phys_to_raw(minimum)}..' \
            f'{phys_to_raw(maximum)} ' \
            f'({round(minimum, 5)}..{round(maximum, 5)} {cg_signal.unit})'
    elif minimum is not None:
        return f'{phys_to_raw(minimum)}.. ({round(minimum, 5)}.. {cg_signal.unit})'
    elif maximum is not None:
        return f'..{phys_to_raw(maximum)} (..{round(maximum, 5)} {cg_signal.unit})'
    else:
        return '-'


def _generate_signal(cg_signal: "CodeGenSignal", bit_fields: bool) -> str:
    comment = _format_comment(cg_signal.signal.comment)
    range_ = _format_range(cg_signal)
    scale = _get(cg_signal.signal.conversion.scale, '-')
    offset = _get(cg_signal.signal.conversion.offset, '-')

    if cg_signal.signal.conversion.is_float or not bit_fields:
        length = ''
    else:
        length = f' : {cg_signal.signal.length}'

    member = SIGNAL_MEMBER_FMT.format(comment=comment,
                                      range=range_,
                                      scale=scale,
                                      offset=offset,
                                      type_name=cg_signal.type_name,
                                      name=cg_signal.snake_name,
                                      length=length)

    return member


def _format_pack_code_mux(cg_message: "CodeGenMessage",
                          mux: SignalTreeMuxElemType,
                          body_lines_per_index: list[str],
                          variable_lines: list[str],
                          helper_kinds: set[THelperKind]) -> list[str]:
    signal_name, multiplexed_signals = next(iter(mux.items()))
    _format_pack_code_signal(cg_message,
                             signal_name,
                             body_lines_per_index,
                             variable_lines,
                             helper_kinds)
    multiplexed_signals_per_id = sorted(multiplexed_signals.items())
    signal_name = camel_to_snake_case(signal_name)

    lines = [
        '',
        f'switch (src_p->{signal_name}) {{'
    ]

    for multiplexer_id, signals_of_multiplexer_id in multiplexed_signals_per_id:
        body_lines = _format_pack_code_level(cg_message,
                                             cast("SignalTreeType", signals_of_multiplexer_id),
                                             variable_lines,
                                             helper_kinds)
        lines.append('')
        lines.append(f'case {multiplexer_id}:')

        if body_lines:
            lines.extend(body_lines[1:-1])

        lines.append('    break;')

    lines.extend([
        '',
        'default:',
        '    break;',
        '}'])

    return [('    ' + line).rstrip() for line in lines]


def _format_pack_code_signal(cg_message: "CodeGenMessage",
                             signal_name: str,
                             body_lines: list[str],
                             variable_lines: list[str],
                             helper_kinds: set[THelperKind]) -> None:
    cg_signal = cg_message.get_signal_by_name(signal_name)

    if cg_signal.signal.conversion.is_float or cg_signal.signal.is_signed:
        variable = f'    uint{cg_signal.type_length}_t {cg_signal.snake_name};'

        if cg_signal.signal.conversion.is_float:
            conversion = f'    memcpy(&{cg_signal.snake_name}, &src_p->{cg_signal.snake_name}, sizeof({cg_signal.snake_name}));'
        else:
            conversion = f'    {cg_signal.snake_name} = (uint{cg_signal.type_length}_t)src_p->{cg_signal.snake_name};'

        variable_lines.append(variable)
        body_lines.append(conversion)

    for index, shift, shift_direction, mask in cg_signal.segments(invert_shift=False):
        if cg_signal.signal.conversion.is_float or cg_signal.signal.is_signed:
            fmt = '    dst_p[{}] |= pack_{}_shift_u{}({}, {}u, 0x{:02x}u);'
        else:
            fmt = '    dst_p[{}] |= pack_{}_shift_u{}(src_p->{}, {}u, 0x{:02x}u);'

        line = fmt.format(index,
                          shift_direction,
                          cg_signal.type_length,
                          cg_signal.snake_name,
                          shift,
                          mask)
        body_lines.append(line)
        helper_kinds.add((shift_direction, cg_signal.type_length))


def _format_pack_code_level(cg_message: "CodeGenMessage",
                            signal_names: SignalTreeType,
                            variable_lines: list[str],
                            helper_kinds: set[THelperKind]) -> list[str]:
    """Format one pack level in a signal tree.

    """

    body_lines: list[str] = []
    muxes_lines: list[str] = []

    for signal_name in signal_names:
        if isinstance(signal_name, dict):
            mux_lines = _format_pack_code_mux(cg_message,
                                              signal_name,
                                              body_lines,
                                              variable_lines,
                                              helper_kinds)
            muxes_lines += mux_lines
        else:
            _format_pack_code_signal(cg_message,
                                     signal_name,
                                     body_lines,
                                     variable_lines,
                                     helper_kinds)

    body_lines = body_lines + muxes_lines

    if body_lines:
        body_lines = ["", *body_lines, ""]

    return body_lines


def _format_pack_code(cg_message: "CodeGenMessage",
                      helper_kinds: set[THelperKind]
                      ) -> tuple[str, str]:
    variable_lines: list[str] = []
    body_lines = _format_pack_code_level(cg_message,
                                         cg_message.message.signal_tree,
                                         variable_lines,
                                         helper_kinds)

    if variable_lines:
        variable_lines = [*sorted(set(variable_lines)), "", ""]

    return '\n'.join(variable_lines), '\n'.join(body_lines)


def _format_unpack_code_mux(cg_message: "CodeGenMessage",
                            mux: SignalTreeMuxElemType,
                            body_lines_per_index: list[str],
                            variable_lines: list[str],
                            helper_kinds: set[THelperKind],
                            node_name: str | None) -> list[str]:
    signal_name, multiplexed_signals = next(iter(mux.items()))
    _format_unpack_code_signal(cg_message,
                               signal_name,
                               body_lines_per_index,
                               variable_lines,
                               helper_kinds)
    multiplexed_signals_per_id = sorted(multiplexed_signals.items())
    signal_name = camel_to_snake_case(signal_name)

    lines = [
        f'switch (dst_p->{signal_name}) {{'
    ]

    for multiplexer_id, signals_of_multiplexer_id in multiplexed_signals_per_id:
        body_lines = _format_unpack_code_level(cg_message,
                                               cast("SignalTreeType", signals_of_multiplexer_id),
                                               variable_lines,
                                               helper_kinds,
                                               node_name)
        lines.append('')
        lines.append(f'case {multiplexer_id}:')
        lines.extend(_strip_blank_lines(body_lines))
        lines.append('    break;')

    lines.extend([
        '',
        'default:',
        '    break;',
        '}'])

    return [('    ' + line).rstrip() for line in lines]


def _format_unpack_code_signal(cg_message: "CodeGenMessage",
                               signal_name: str,
                               body_lines: list[str],
                               variable_lines: list[str],
                               helper_kinds: set[THelperKind]) -> None:
    cg_signal = cg_message.get_signal_by_name(signal_name)
    conversion_type_name = f'uint{cg_signal.type_length}_t'

    if cg_signal.signal.conversion.is_float or cg_signal.signal.is_signed:
        variable = f'    {conversion_type_name} {cg_signal.snake_name};'
        variable_lines.append(variable)

    segments = cg_signal.segments(invert_shift=True)

    for i, (index, shift, shift_direction, mask) in enumerate(segments):
        if cg_signal.signal.conversion.is_float or cg_signal.signal.is_signed:
            fmt = '    {} {} unpack_{}_shift_u{}(src_p[{}], {}u, 0x{:02x}u);'
        else:
            fmt = '    dst_p->{} {} unpack_{}_shift_u{}(src_p[{}], {}u, 0x{:02x}u);'

        line = fmt.format(cg_signal.snake_name,
                          '=' if i == 0 else '|=',
                          shift_direction,
                          cg_signal.type_length,
                          index,
                          shift,
                          mask)
        body_lines.append(line)
        helper_kinds.add((shift_direction, cg_signal.type_length))

    if cg_signal.signal.conversion.is_float:
        conversion = f'    memcpy(&dst_p->{cg_signal.snake_name}, &{cg_signal.snake_name}, sizeof(dst_p->{cg_signal.snake_name}));'
        body_lines.append(conversion)
    elif cg_signal.signal.is_signed:
        mask = ((1 << (cg_signal.type_length - cg_signal.signal.length)) - 1)

        if mask != 0:
            mask <<= cg_signal.signal.length
            formatted = SIGN_EXTENSION_FMT.format(name=cg_signal.snake_name,
                                                  shift=cg_signal.signal.length - 1,
                                                  mask=mask,
                                                  suffix=cg_signal.conversion_type_suffix)
            body_lines.extend(formatted.splitlines())

        conversion = f'    dst_p->{cg_signal.snake_name} = (int{cg_signal.type_length}_t){cg_signal.snake_name};'
        body_lines.append(conversion)


def _format_unpack_code_level(cg_message: "CodeGenMessage",
                              signal_names: SignalTreeType,
                              variable_lines: list[str],
                              helper_kinds: set[THelperKind],
                              node_name: str | None) -> list[str]:
    """Format one unpack level in a signal tree.

    """

    body_lines: list[str] = []
    muxes_lines: list[str] = []

    for signal_name in signal_names:
        if isinstance(signal_name, dict):
            mux_lines = _format_unpack_code_mux(cg_message,
                                                signal_name,
                                                body_lines,
                                                variable_lines,
                                                helper_kinds,
                                                node_name)

            if muxes_lines:
                muxes_lines.append('')

            muxes_lines += mux_lines
        else:
            if not _is_receiver(cg_message.get_signal_by_name(signal_name), node_name):
                continue

            _format_unpack_code_signal(cg_message,
                                       signal_name,
                                       body_lines,
                                       variable_lines,
                                       helper_kinds)

    if body_lines and body_lines[-1] != '':
        body_lines.append('')

    if muxes_lines:
        muxes_lines.append('')

    body_lines = body_lines + muxes_lines

    if body_lines:
        body_lines = ["", *body_lines]

    return body_lines


def _format_unpack_code(cg_message: "CodeGenMessage",
                        helper_kinds: set[THelperKind],
                        node_name: str | None) -> tuple[str, str]:
    variable_lines: list[str] = []
    body_lines = _format_unpack_code_level(cg_message,
                                           cg_message.message.signal_tree,
                                           variable_lines,
                                           helper_kinds,
                                           node_name)

    if variable_lines:
        variable_lines = [*sorted(set(variable_lines)), "", ""]

    return '\n'.join(variable_lines), '\n'.join(body_lines)


def _generate_struct(cg_message: "CodeGenMessage", bit_fields: bool) -> tuple[str, list[str]]:
    members = []

    for cg_signal in cg_message.cg_signals:
        members.append(_generate_signal(cg_signal, bit_fields))

    if not members:
        members = [
            ('    /**\n'
            '     * Dummy signal in empty message.\n'
            '     */\n'
            '    uint8_t dummy;')
        ]

    if cg_message.message.comment is None:
        comment = ''
    else:
        comment = f' * {cg_message.message.comment}\n *\n'

    return comment, members


def _format_choices(cg_signal: "CodeGenSignal", signal_name: str) -> list[str]:
    choices = []

    for value, name in sorted(cg_signal.unique_choices.items()):
        if cg_signal.signal.is_signed:
            fmt = '{signal_name}_{name}_CHOICE ({value})'
        else:
            fmt = '{signal_name}_{name}_CHOICE ({value}u)'

        choices.append(fmt.format(signal_name=signal_name.upper(),
                                  name=str(name),
                                  value=value))

    return choices


def _generate_encode_decode(cg_signal: "CodeGenSignal", use_float: bool, use_round: bool) -> tuple[str, str]:
    floating_point_type = _get_floating_point_type(use_float)

    scale = cg_signal.signal.scale
    offset = cg_signal.signal.offset

    scale_literal = f"{scale}{'.0' if isinstance(scale, int) else ''}{'f' if use_float else ''}"
    offset_literal = f"{offset}{'.0' if isinstance(offset, int) else ''}{'f' if use_float else ''}"

    if offset == 0 and scale == 1:
        encoding = 'value'
        decoding = f'({floating_point_type})value'
    elif offset != 0 and scale != 1:
        encoding = f'(value - {offset_literal}) / {scale_literal}'
        decoding = f'(({floating_point_type})value * {scale_literal}) + {offset_literal}'
    elif offset != 0:
        encoding = f'value - {offset_literal}'
        decoding = f'({floating_point_type})value + {offset_literal}'
    else:
        encoding = f'value / {scale_literal}'
        decoding = f'({floating_point_type})value * {scale_literal}'

    if not cg_signal.signal.is_float and use_round:
        encoding = f'round{"f" if use_float else ""}({encoding})'

    return encoding, decoding


def _generate_is_in_range(cg_signal: "CodeGenSignal") -> str:
    """Generate range checks for all signals in given message.

    """
    minimum = cg_signal.signal.minimum
    maximum = cg_signal.signal.maximum

    if minimum is not None:
        minimum = cg_signal.signal.scaled_to_raw(minimum)

    if maximum is not None:
        maximum = cg_signal.signal.scaled_to_raw(maximum)

    if (minimum is None and cg_signal.minimum_can_raw_value is not None) and \
       (cg_signal.minimum_ctype_value is None or cg_signal.minimum_can_raw_value > cg_signal.minimum_ctype_value):
        minimum = cg_signal.minimum_can_raw_value

    if (maximum is None and cg_signal.maximum_can_raw_value is not None) and \
       (cg_signal.maximum_ctype_value is None or cg_signal.maximum_can_raw_value < cg_signal.maximum_ctype_value):
        maximum = cg_signal.maximum_can_raw_value

    suffix = cg_signal.type_suffix
    check = []

    if minimum is not None:
        if not cg_signal.signal.conversion.is_float:
            minimum = round(minimum)
        else:
            minimum = float(minimum)

        minimum_ctype_value = cg_signal.minimum_ctype_value

        if (minimum_ctype_value is None) or (minimum > minimum_ctype_value):
            check.append(f'(value >= {minimum}{suffix})')

    if maximum is not None:
        if not cg_signal.signal.conversion.is_float:
            maximum = round(maximum)
        else:
            maximum = float(maximum)

        maximum_ctype_value = cg_signal.maximum_ctype_value

        if (maximum_ctype_value is None) or (maximum < maximum_ctype_value):
            check.append(f'(value <= {maximum}{suffix})')

    if not check:
        check = ['true']
    elif len(check) == 1:
        check = [check[0][1:-1]]

    return ' && '.join(check)

def _generate_is_in_phys_range(cg_signal: "CodeGenSignal", use_float: bool) -> tuple[str, str]:
    """Generate is_in_range check in the physical (floating-point) domain.

    The standard is_in_range() checks the raw integer value AFTER encode().
    If the physical value overflows the raw type, it can produce a raw value
    that wrongly passes is_in_range() — silent data corruption.

    Example — uint8 signal, scale=1, offset=0, range [0, 100]:
        encode(260.0) -> (uint8_t)(260) = 4   -> is_in_range(4)   -> TRUE  <- BUG
        encode(-1.0)  -> (uint8_t)(-1)  = 255 -> is_in_range(255) -> FALSE <- ok by chance
        is_in_phys_range(260.0) -> FALSE  <- correct
        is_in_phys_range(-1.0)  -> FALSE  <- correct

    Returns:
        phys_check  -- C expression for the return statement
        phys_unused -- (void)value cast if no bounds defined, else ""
    """
    minimum = cg_signal.signal.minimum
    maximum = cg_signal.signal.maximum
    f_suffix = "f" if use_float else ""
    check = []

    if minimum is not None:
        check.append(f"(value >= {float(minimum)}{f_suffix})")

    if maximum is not None:
        check.append(f"(value <= {float(maximum)}{f_suffix})")

    if not check:
        return "true", "    (void)value;\n\n"
    elif len(check) == 1:
        return check[0][1:-1], ""
    else:
        return " && ".join(check), ""

def _generate_frame_id_defines(database_name: str,
                               cg_messages: list["CodeGenMessage"],
                               node_name: str | None) -> str:
    return '\n'.join([
        f'#define {database_name.upper()}_{cg_message.snake_name.upper()}_FRAME_ID (0x{cg_message.message.frame_id:02x}u)'
        for cg_message in cg_messages if _is_sender_or_receiver(cg_message, node_name)
    ])


def _generate_frame_length_defines(database_name: str,
                                   cg_messages: list["CodeGenMessage"],
                                   node_name: str | None) -> str:
    result = '\n'.join([
        f'#define {database_name.upper()}_{cg_message.snake_name.upper()}_LENGTH ({cg_message.message.length}u)'
        for cg_message in cg_messages if _is_sender_or_receiver(cg_message, node_name)
    ])

    return result


def _generate_frame_cycle_time_defines(database_name: str,
                                       cg_messages: list["CodeGenMessage"],
                                       node_name: str | None) -> str:
    result = '\n'.join([
        f'#define {database_name.upper()}_{cg_message.snake_name.upper()}_CYCLE_TIME_MS ({cg_message.message.cycle_time}u)'
        for cg_message in cg_messages if cg_message.message.cycle_time is not None and
                                      _is_sender_or_receiver(cg_message, node_name)
    ])

    return result


def _generate_is_extended_frame_defines(database_name: str,
                                        cg_messages: list["CodeGenMessage"],
                                        node_name: str | None) -> str:
    result = '\n'.join([
        f'#define {database_name.upper()}_{cg_message.snake_name.upper()}_IS_EXTENDED ({int(cg_message.message.is_extended_frame)})'
        for cg_message in cg_messages if _is_sender_or_receiver(cg_message, node_name)
    ])

    return result


def _generate_choices_defines(database_name: str,
                              cg_messages: list["CodeGenMessage"],
                              node_name: str | None) -> str:
    choices_defines = []

    for cg_message in cg_messages:
        is_sender = _is_sender(cg_message, node_name)
        for cg_signal in cg_message.cg_signals:
            if cg_signal.signal.conversion.choices is None:
                continue
            if not is_sender and not _is_receiver(cg_signal, node_name):
                continue

            choices = _format_choices(cg_signal, cg_signal.snake_name)
            signal_choices_defines = '\n'.join([
                f'#define {database_name.upper()}_{cg_message.snake_name.upper()}_{choice}'
                for choice in choices
            ])
            choices_defines.append(signal_choices_defines)

    return '\n\n'.join(choices_defines)


def _generate_frame_name_macros(database_name: str,
                                cg_messages: list["CodeGenMessage"],
                                node_name: str | None) -> str:
    result = '\n'.join([
        f'#define {database_name.upper()}_{cg_message.snake_name.upper()}_NAME "{cg_message.message.name}"'
        for cg_message in cg_messages if _is_sender_or_receiver(cg_message, node_name)
    ])

    return result


def _generate_signal_name_macros(database_name: str,
                                 cg_messages: list["CodeGenMessage"],
                                 node_name: str | None) -> str:
    result = '\n'.join([
        f'#define {database_name.upper()}_{cg_message.snake_name.upper()}_{cg_signal.snake_name.upper()}_NAME "{cg_signal.signal.name}"'
        for cg_message in cg_messages if _is_sender_or_receiver(cg_message, node_name) for cg_signal in cg_message.cg_signals
    ])

    return result


def _generate_structs(database_name: str,
                      cg_messages: list["CodeGenMessage"],
                      bit_fields: bool,
                      node_name: str | None) -> str:
    structs = []

    for cg_message in cg_messages:
        if _is_sender_or_receiver(cg_message, node_name):
            comment, members = _generate_struct(cg_message, bit_fields)
            structs.append(
                STRUCT_FMT.format(comment=comment,
                                  database_message_name=cg_message.message.name,
                                  message_name=cg_message.snake_name,
                                  database_name=database_name,
                                  members='\n\n'.join(members)))

    return '\n'.join(structs)


def _is_sender(cg_message: "CodeGenMessage", node_name: str | None) -> bool:
    return node_name is None or node_name in cg_message.message.senders


def _is_receiver(cg_signal: "CodeGenSignal", node_name: str | None) -> bool:
    return node_name is None or node_name in cg_signal.signal.receivers


def _is_sender_or_receiver(cg_message: "CodeGenMessage", node_name: str | None) -> bool:
    if _is_sender(cg_message, node_name):
        return True
    return any(_is_receiver(cg_signal, node_name) for cg_signal in cg_message.cg_signals)


def _get_floating_point_type(use_float: bool) -> str:
    return 'float' if use_float else 'double'


def _generate_declarations(database_name: str,
                           cg_messages: list["CodeGenMessage"],
                           floating_point_numbers: bool,
                           use_float: bool,
                           node_name: str | None) -> str:
    declarations = []

    for cg_message in cg_messages:
        signal_declarations = []
        is_sender = _is_sender(cg_message, node_name)
        is_receiver = node_name is None

        for cg_signal in cg_message.cg_signals:
            if _is_receiver(cg_signal, node_name):
                is_receiver = True

            signal_declaration = ''

            if floating_point_numbers:
                if is_sender:
                    signal_declaration += SIGNAL_DECLARATION_ENCODE_FMT.format(
                        database_name=database_name,
                        message_name=cg_message.snake_name,
                        signal_name=cg_signal.snake_name,
                        type_name=cg_signal.type_name,
                        floating_point_type=_get_floating_point_type(use_float))
                if node_name is None or _is_receiver(cg_signal, node_name):
                    signal_declaration += SIGNAL_DECLARATION_DECODE_FMT.format(
                        database_name=database_name,
                        message_name=cg_message.snake_name,
                        signal_name=cg_signal.snake_name,
                        type_name=cg_signal.type_name,
                        floating_point_type=_get_floating_point_type(use_float))

            if is_sender or _is_receiver(cg_signal, node_name):
                signal_declaration += SIGNAL_DECLARATION_IS_IN_RANGE_FMT.format(
                    database_name=database_name,
                    message_name=cg_message.snake_name,
                    signal_name=cg_signal.snake_name,
                    type_name=cg_signal.type_name)

                signal_declaration += SIGNAL_DECLARATION_IS_IN_PHYS_RANGE_FMT.format(
                    database_name=database_name,
                    message_name=cg_message.snake_name,
                    signal_name=cg_signal.snake_name,
                    floating_point_type=_get_floating_point_type(use_float))

                signal_declarations.append(signal_declaration)
        declaration = ""
        if is_sender:
            declaration += DECLARATION_PACK_FMT.format(database_name=database_name,
                                                       database_message_name=cg_message.message.name,
                                                       message_name=cg_message.snake_name)
        if is_receiver:
            declaration += DECLARATION_UNPACK_FMT.format(database_name=database_name,
                                                         database_message_name=cg_message.message.name,
                                                         message_name=cg_message.snake_name)

        if is_sender or is_receiver:
            declaration += MESSAGE_DECLARATION_INIT_FMT.format(database_name=database_name,
                                                            database_message_name=cg_message.message.name,
                                                            message_name=cg_message.snake_name)

        if signal_declarations:
            declaration += '\n' + '\n'.join(signal_declarations)

        if declaration:
            declarations.append(declaration)

    return '\n'.join(declarations)


def _generate_definitions(database_name: str,
                          cg_messages: list["CodeGenMessage"],
                          floating_point_numbers: bool,
                          use_float: bool,
                          node_name: str | None,
                          use_round: bool,
                          ) -> tuple[str, tuple[set[THelperKind], set[THelperKind]]]:
    definitions = []
    pack_helper_kinds: set[THelperKind] = set()
    unpack_helper_kinds: set[THelperKind] = set()

    for cg_message in cg_messages:
        signal_definitions = []
        is_sender = _is_sender(cg_message, node_name)
        is_receiver = node_name is None
        signals_init_body = ''

        for cg_signal in cg_message.cg_signals:
            if use_float and cg_signal.type_name == "double":
                warnings.warn(f"User selected `--use-float`, but database contains "
                              f"signal with data type `double`: "
                              f"\"{cg_message.message.name}::{cg_signal.signal.name}\"",
                              stacklevel=2)
                _use_float = False
            else:
                _use_float = use_float

            encode, decode = _generate_encode_decode(cg_signal, _use_float, use_round)
            check = _generate_is_in_range(cg_signal)
            phys_check, phys_unused = _generate_is_in_phys_range(cg_signal, _use_float)

            if _is_receiver(cg_signal, node_name):
                is_receiver = True

            if check == 'true':
                unused = '    (void)value;\n\n'
            else:
                unused = ''

            signal_definition = ''

            if floating_point_numbers:
                if is_sender:
                    signal_definition += SIGNAL_DEFINITION_ENCODE_FMT.format(
                        database_name=database_name,
                        message_name=cg_message.snake_name,
                        signal_name=cg_signal.snake_name,
                        type_name=cg_signal.type_name,
                        encode=encode,
                        floating_point_type=_get_floating_point_type(_use_float))
                if node_name is None or _is_receiver(cg_signal, node_name):
                    signal_definition += SIGNAL_DEFINITION_DECODE_FMT.format(
                        database_name=database_name,
                        message_name=cg_message.snake_name,
                        signal_name=cg_signal.snake_name,
                        type_name=cg_signal.type_name,
                        decode=decode,
                        floating_point_type=_get_floating_point_type(_use_float))

            if is_sender or _is_receiver(cg_signal, node_name):
                signal_definition += SIGNAL_DEFINITION_IS_IN_RANGE_FMT.format(
                    database_name=database_name,
                    message_name=cg_message.snake_name,
                    signal_name=cg_signal.snake_name,
                    type_name=cg_signal.type_name,
                    unused=unused,
                    check=check)

                signal_definition += SIGNAL_DEFINITION_IS_IN_PHYS_RANGE_FMT.format(
                    database_name=database_name,
                    message_name=cg_message.snake_name,
                    signal_name=cg_signal.snake_name,
                    floating_point_type=_get_floating_point_type(_use_float),
                    phys_check=phys_check,
                    phys_unused=phys_unused)

                signal_definitions.append(signal_definition)

            if cg_signal.signal.initial:
                signals_init_body += INIT_SIGNAL_BODY_TEMPLATE_FMT.format(signal_initial=cg_signal.signal.raw_initial,
                                                                          signal_name=cg_signal.snake_name)

        if cg_message.message.length > 0:
            pack_variables, pack_body = _format_pack_code(cg_message,
                                                          pack_helper_kinds)
            unpack_variables, unpack_body = _format_unpack_code(cg_message,
                                                                unpack_helper_kinds,
                                                                node_name)
            pack_unused = ''
            unpack_unused = ''

            if not pack_body:
                pack_unused += '    (void)src_p;\n\n'

            if not unpack_body:
                unpack_unused += '    (void)dst_p;\n'
                unpack_unused += '    (void)src_p;\n\n'

            definition = ""
            if is_sender:
                definition += DEFINITION_PACK_FMT.format(database_name=database_name,
                                                         database_message_name=cg_message.message.name,
                                                         message_name=cg_message.snake_name,
                                                         message_length=cg_message.message.length,
                                                         pack_unused=pack_unused,
                                                         pack_variables=pack_variables,
                                                         pack_body=pack_body)
            if is_receiver:
                definition += DEFINITION_UNPACK_FMT.format(database_name=database_name,
                                                           database_message_name=cg_message.message.name,
                                                           message_name=cg_message.snake_name,
                                                           message_length=cg_message.message.length,
                                                           unpack_unused=unpack_unused,
                                                           unpack_variables=unpack_variables,
                                                           unpack_body=unpack_body)

            if is_sender or is_receiver:
                definition += MESSAGE_DEFINITION_INIT_FMT.format(database_name=database_name,
                                                                database_message_name=cg_message.message.name,
                                                                message_name=cg_message.snake_name,
                                                                init_body=signals_init_body)

        else:
            definition = EMPTY_DEFINITION_FMT.format(database_name=database_name,
                                                     message_name=cg_message.snake_name)

        if signal_definitions:
            definition += '\n' + '\n'.join(signal_definitions)

        if definition:
            definitions.append(definition)

    return '\n'.join(definitions), (pack_helper_kinds, unpack_helper_kinds)


def _generate_helpers_kind(kinds: set[THelperKind],
                           left_format: str,
                           right_format: str) -> list[str]:
    formats = {
        'left': left_format,
        'right': right_format
    }
    helpers = []

    for shift_direction, length in sorted(kinds):
        var_type = f'uint{length}_t'
        helper = formats[shift_direction].format(length=length,
                                                 var_type=var_type)
        helpers.append(helper)

    return helpers


def _generate_helpers(kinds: tuple[set[THelperKind], set[THelperKind]]) -> str:
    pack_helpers = _generate_helpers_kind(kinds[0],
                                          PACK_HELPER_LEFT_SHIFT_FMT,
                                          PACK_HELPER_RIGHT_SHIFT_FMT)
    unpack_helpers = _generate_helpers_kind(kinds[1],
                                            UNPACK_HELPER_LEFT_SHIFT_FMT,
                                            UNPACK_HELPER_RIGHT_SHIFT_FMT)
    helpers = pack_helpers + unpack_helpers

    if helpers:
        helpers.append('')

    return '\n'.join(helpers)


def _generate_fuzzer_source(database_name: str,
                            cg_messages: list["CodeGenMessage"],
                            date: str,
                            header_name: str,
                            source_name: str,
                            fuzzer_source_name: str) -> tuple[str, str]:
    tests = []
    calls = []

    for cg_message in cg_messages:
        name = f'{database_name}_{camel_to_snake_case(cg_message.message.name)}'

        test = TEST_FMT.format(name=name)
        tests.append(test)

        call = f'    test_{name}(data_p, size);'
        calls.append(call)

    source = FUZZER_SOURCE_FMT.format(version=__version__,
                                      date=date,
                                      header=header_name,
                                      tests='\n'.join(tests),
                                      llvm_body='\n'.join(calls))

    makefile = FUZZER_MAKEFILE_FMT.format(version=__version__,
                                          date=date,
                                          source=source_name,
                                          fuzzer_source=fuzzer_source_name)

    return source, makefile


def generate(database: "Database",
             database_name: str,
             header_name: str,
             source_name: str,
             fuzzer_source_name: str,
             floating_point_numbers: bool = True,
             bit_fields: bool = False,
             use_float: bool = False,
             node_name: str | None = None,
             use_round: bool = False,
             ) -> tuple[str, str, str, str]:
    """Generate C source code from given CAN database `database`.

    `database_name` is used as a prefix for all defines, data
    structures and functions.

    `header_name` is the file name of the C header file, which is
    included by the C source file.

    `source_name` is the file name of the C source file, which is
    needed by the fuzzer makefile.

    `fuzzer_source_name` is the file name of the C source file, which
    is needed by the fuzzer makefile.

    Set `floating_point_numbers` to ``True`` to allow floating point
    numbers in the generated code.

    Set `bit_fields` to ``True`` to generate bit fields in structs.

    Set `use_float` to ``True`` to prefer the `float` type instead
    of the `double` type for floating point numbers.

    `node_name` specifies the node for which message packers will be generated.
    For all other messages, unpackers will be generated. If `node_name` is not
    provided, both packers and unpackers will be generated.

    Set `use_round` to ``True`` to round encoded values to the nearest
    integer (instead of truncating).

    This function returns a tuple of the C header and source files as
    strings.

    """

    date = time.ctime()
    cg_messages = [CodeGenMessage(message) for message in database.messages]
    include_guard = f'{database_name.upper()}_H'
    frame_id_defines = _generate_frame_id_defines(database_name, cg_messages, node_name)
    frame_length_defines = _generate_frame_length_defines(database_name,
                                                          cg_messages,
                                                          node_name)
    is_extended_frame_defines = _generate_is_extended_frame_defines(
        database_name,
        cg_messages,
        node_name)
    frame_cycle_time_defines = _generate_frame_cycle_time_defines(
        database_name,
        cg_messages,
        node_name)
    choices_defines = _generate_choices_defines(database_name, cg_messages, node_name)

    frame_name_macros = _generate_frame_name_macros(database_name, cg_messages, node_name)
    signal_name_macros = _generate_signal_name_macros(database_name, cg_messages, node_name)

    structs = _generate_structs(database_name, cg_messages, bit_fields, node_name)
    declarations = _generate_declarations(database_name,
                                          cg_messages,
                                          floating_point_numbers,
                                          use_float,
                                          node_name)
    definitions, helper_kinds = _generate_definitions(database_name,
                                                      cg_messages,
                                                      floating_point_numbers,
                                                      use_float,
                                                      node_name,
                                                      use_round)
    helpers = _generate_helpers(helper_kinds)

    header = HEADER_FMT.format(file_name=header_name,
                               version=__version__,
                               date=date,
                               include_guard=include_guard,
                               frame_id_defines=frame_id_defines,
                               frame_length_defines=frame_length_defines,
                               is_extended_frame_defines=is_extended_frame_defines,
                               frame_cycle_time_defines=frame_cycle_time_defines,
                               choices_defines=choices_defines,
                               frame_name_macros=frame_name_macros,
                               signal_name_macros=signal_name_macros,
                               structs=structs,
                               declarations=declarations)

    includes = []
    if use_round:
        includes.append('#include <math.h>\n')

    source = SOURCE_FMT.format(file_name=source_name,
                               version=__version__,
                               date=date,
                               header=header_name,
                               helpers=helpers,
                               definitions=definitions,
                               includes=''.join(includes))

    fuzzer_source, fuzzer_makefile = _generate_fuzzer_source(
        database_name,
        cg_messages,
        date,
        header_name,
        source_name,
        fuzzer_source_name)

    return header, source, fuzzer_source, fuzzer_makefile



# utility functions that are helpful when dealing with ARXML files


def parse_number_string(in_string: str, allow_float: bool=False) \
    -> int | float:
    """Convert a string representing numeric value that is specified
    within an ARXML file to either an integer or a floating point object

    This is surprisingly complicated:

    - Some ARXML files use "true" and "false" synonymous to 1 and 0
    - ARXML uses the C notation (leading 0) to specify octal numbers
      whereas python only accepts the "0o" prefix
    - Some ARXML editors seem to sometimes include a dot in integer
      numbers (e.g., they produce "123.0" instead of "123")
    """
    ret: int | float | None = None
    in_string = in_string.strip().lower()

    if len(in_string) > 0:
        # the string literals "true" and "false" are interpreted as 1 and 0
        if in_string == 'true':
            ret = 1

        if in_string == 'false':
            ret = 0

        # note: prefer parsing as integer first to prevent floating-point precision issues in large numbers.
        # 1. try int parsing from octal notation without an "o" after the leading 0.
        if len(in_string) > 1 and in_string[0] == '0' and in_string[1].isdigit():
            # interpret strings starting with a 0 as octal because
            # python's int(*, 0) does not for some reason.
            ret = int(in_string, 8)

        # 2. try int parsing with auto-detected base.
        if ret is None:
            # handles python integer literals
            # see https://docs.python.org/3/reference/lexical_analysis.html#integers
            try:
                ret = int(in_string, 0)
            except ValueError:
                pass

        # 3. try float parsing from hex string.
        if ret is None and in_string.startswith('0x'):
            ret = float.fromhex(in_string)

        # 4. try float parsing from 'normal' float string
        if ret is None:
            # throws an error, if non-numeric
            # but handles for example scientific notation
            ret = float(in_string)

        # check for not allowed non-integer values
        if not allow_float:
            if ret != int(ret):
                raise ValueError('Floating point value specified where integer '
                                 'is required')
            # if an integer is required but a .0 floating point value is
            # specified, we accept the input anyway. (this seems to be an
            # ambiguity in the AUTOSAR specification.)
            ret = int(ret)
    else:
        ret = 0

    return ret




import argparse
import os
import sys

from ... import database
from ...database import Message
from ...database.can.database import Database as CanDatabase
from ...database.diagnostics.database import Database as DiagnosticsDatabase
from ...database.utils import format_and
from ...j1939 import frame_id_unpack, is_pdu_format_1, pgn_pack
from . import formatting


def _print_j1939_frame_id(message):
    unpacked = frame_id_unpack(message.frame_id)

    print(f'      Priority:       {unpacked.priority}')

    if is_pdu_format_1(unpacked.pdu_format):
        pdu_format = 'PDU 1'
        pdu_specific = 0
        destination = f'0x{unpacked.pdu_specific:02x}'
    else:
        pdu_format = 'PDU 2'
        pdu_specific = unpacked.pdu_specific
        destination = 'All'

    print('      PGN:            0x{:05x}'.format(
        pgn_pack(unpacked.reserved,
                 unpacked.data_page,
                 unpacked.pdu_format,
                 pdu_specific)))
    print(f'      Source:         0x{unpacked.source_address:02x}')
    print(f'      Destination:    {destination}')
    print(f'      Format:         {pdu_format}')


def _dump_can_message(message, with_comments=False, name_prefix='', WIDTH=None):
    cycle_time = message.cycle_time
    signal_choices_string = formatting.signal_choices_string(message)

    if cycle_time is None:
        cycle_time = '-'

    if len(message.senders) == 0:
        message.senders.append('-')

    print()
    print(f'  Name:           {name_prefix}{message.name}')
    if message.frame_id is not None and not name_prefix:
        # only print the arbitration ID for top-level messages
        print(f'  Id:             0x{message.frame_id:x}')
    if message.header_id is not None and name_prefix:
        # only print the header ID for child messages
        print(f'  Header id:      0x{message._header_id:06x}')

    if message.protocol == 'j1939':
        _print_j1939_frame_id(message)

    if message.is_container:
        print(f'  Maximum length: {message.length} bytes')
    else:
        print(f'  Length:         {message.length} bytes')

    print(f'  Cycle time:     {cycle_time} ms')
    print(f'  Senders:        {format_and(message.senders)}')
    if message.is_container:
        print('  Possibly contained children:')
        print()
        for child in message.contained_messages:
            print(f'      {message.name} :: {child.name}')
        print()
    else:
        print('  Layout:')
        print()
        print('\n'.join([
            ('    ' + line).rstrip()
            for line in formatting.layout_string(message).splitlines()
        ]))
        print()
        print('  Signal tree:')
        print()
        print('\n'.join([
            ('    ' + line).rstrip()
            for line in formatting.signal_tree_string(message, WIDTH, with_comments=with_comments).splitlines()
        ]))
        print()

        if signal_choices_string:
            print('  Signal choices:')
            print('\n'.join([
                ('    ' + line).rstrip()
                for line in signal_choices_string.splitlines()
            ]))
            print()

    print('  ' + 72 * '-')

    if message.is_container:
        # dump the layout of the child messages of the container
        for child in message.contained_messages:
            _dump_can_message(child,
                              with_comments=with_comments,
                              WIDTH=WIDTH,
                              name_prefix=f'{message.name} :: ')


def _dump_can_messages(msg_list: list[Message], with_comments: bool = False) -> None:
    WIDTH = 80
    try:
        WIDTH, _ = os.get_terminal_size()
    except OSError:
        pass

    print('================================= Messages =================================')
    print()
    print('  ' + 72 * '-')

    for message in msg_list:
        _dump_can_message(message,
                          with_comments=with_comments,
                          WIDTH=WIDTH)


def _dump_diagnostics_database(dbase):
    print('=================================== Dids ===================================')
    print()
    print('  ' + 72 * '-')

    for did in dbase.dids:
        print()
        print(f'  Name:       {did.name}')
        print(f'  Length:     {did.length} bytes')
        print('  Layout:')
        print()

        for data in did.datas:
            print(f'    Name:      {data.name}')
            print(f'    Start bit: {data.start}')
            print(f'    Length:    {data.length}')
            print()

        print()
        print('  ' + 72 * '-')


def _do_dump(args):
    dbase = database.load_file(args.database,
                               encoding=args.encoding,
                               prune_choices=args.prune,
                               strict=not args.no_strict)
    message_name = args.message
    if isinstance(dbase, CanDatabase):
        if message_name is None:
            _dump_can_messages(dbase.messages, args.with_comments)
        else:
            try:
                message = dbase.get_message_by_name(message_name)
            except KeyError:
                raise KeyError(f'Unknown message {message_name}') from KeyError
            _dump_can_messages([message], args.with_comments)

    elif isinstance(dbase, DiagnosticsDatabase):
        _dump_diagnostics_database(dbase)
    else:
        sys.exit('Unsupported database type.')


def add_subparser(subparsers):
    dump_parser = subparsers.add_parser(
        'dump',
        description='Dump given database in a human readable format.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    dump_parser.add_argument(
        '-e', '--encoding',
        help='File encoding.')
    dump_parser.add_argument(
        '--prune',
        action='store_true',
        help='Try to shorten the names of named signal choices.')
    dump_parser.add_argument(
        '--no-strict',
        action='store_true',
        help='Skip database consistency checks.')
    dump_parser.add_argument(
        'database',
        help='Database file.')
    dump_parser.add_argument(
        '--with-comments',
        action='store_true',
        help='Print the comments.')
    dump_parser.add_argument(
        '-m', '--message',
        help='Print only the specified message.')
    dump_parser.set_defaults(func=_do_dump)



from collections.abc import Iterable

from cantools.database.errors import DecodeError

from ..database.can.database import Database
from ..database.can.message import Message
from ..database.namedsignalvalue import NamedSignalValue
from ..typechecking import (
    ContainerDecodeResultType,
    ContainerUnpackResultType,
    SignalDictType,
)

MULTI_LINE_FMT = '''
{message}(
{signals}
)\
'''


def format_signals(message, decoded_signals):
    formatted_signals = []

    for signal in message.signals:
        try:
            value = decoded_signals[signal.name]
        except KeyError:
            continue

        signal_name = signal.name

        if signal.unit is None or \
           isinstance(value, (NamedSignalValue, str)):

            formatted_signal = f'{signal_name}: {value}'

        else:
            formatted_signal = f'{signal_name}: {value} {signal.unit}'

        formatted_signals.append(formatted_signal)

    return formatted_signals


def _format_message_single_line(name : str,
                                formatted_signals : Iterable[str]) -> str:
    return ' {}({})'.format(name,
                            ', '.join(formatted_signals))


def _format_message_multi_line(name: str,
                               formatted_signals : Iterable[str]) -> str:
    indented_signals = [
        '    ' + formatted_signal
        for formatted_signal in formatted_signals
    ]

    return MULTI_LINE_FMT.format(message=name,
                                 signals=',\n'.join(indented_signals))

def _format_container_single_line(message : Message,
                                  unpacked_data : ContainerUnpackResultType,
                                  decoded_data : ContainerDecodeResultType) \
                                  -> str:
    contained_list = []
    for i, (cm, signals) in enumerate(decoded_data):
        if isinstance(cm, Message):
            if isinstance(signals, bytes):
                formatted_cm = f'{cm.name}: Undecodable data: {signals.hex(" ")}'
            else:
                formatted_cm_signals = format_signals(cm, signals)
                formatted_cm = _format_message_single_line(cm.name, formatted_cm_signals)
            contained_list.append(formatted_cm)
        else:
            header_id = cm
            data = unpacked_data[i][1]
            contained_list.append(
                f'(Unknown contained message: Header ID: 0x{header_id:x}, '
                f'Data: {bytes(data).hex()})')

    return f' {message.name}({", ".join(contained_list)})'


def _format_container_multi_line(message : Message,
                                 unpacked_data : ContainerUnpackResultType,
                                 decoded_data : ContainerDecodeResultType) -> str:
    contained_list = []
    for i, (cm, signals) in enumerate(decoded_data):
        if isinstance(cm, Message):
            if isinstance(signals, bytes):
                formatted_cm = f'    {cm.header_id:06x}##{signals.hex()} ::\n'
                formatted_cm += f'    {cm.name}: Undecodable data'
                contained_list.append(formatted_cm)
            else:
                formatted_cm_signals = format_signals(cm, signals)
                formatted_cm = f'{cm.header_id:06x}##'
                formatted_cm += f'{bytes(unpacked_data[i][1]).hex()} ::'
                formatted_cm += _format_message_multi_line(cm.name, formatted_cm_signals)
                formatted_cm = formatted_cm.replace('\n', '\n    ')
                contained_list.append('    '+formatted_cm.strip())
        else:
            header_id = cm
            data = unpacked_data[i][1]
            contained_list.append(
                f'    {header_id:06x}##{data.hex()} ::\n'
                f'    Unknown contained message')

    return \
        f'\n{message.name}(\n' + \
        ',\n'.join(contained_list) + \
        '\n)'

def format_message_by_frame_id(dbase : Database,
                               frame_id : int,
                               data : bytes,
                               decode_choices : bool,
                               single_line : bool,
                               decode_containers : bool,
                               *,
                               allow_truncated: bool,
                               allow_excess: bool) -> str:
    try:
        message = dbase.get_message_by_frame_id(frame_id)
    except KeyError:
        return f' Unknown frame id {frame_id} (0x{frame_id:x})'

    if message.is_container:
        if decode_containers:
            return format_container_message(message,
                                            data,
                                            decode_choices,
                                            single_line,
                                            allow_truncated=allow_truncated,
                                            allow_excess=allow_excess)
        else:
            return f' Frame 0x{frame_id:x} is a container message'

    try:
        decoded_signals = message.decode_simple(data,
                                        decode_choices,
                                        allow_truncated=allow_truncated,
                                        allow_excess=allow_excess)

        return format_message(message, decoded_signals, single_line)
    except DecodeError as e:
        return f' {e}'

def format_container_message(message : Message,
                             data : bytes,
                             decode_choices : bool,
                             single_line : bool,
                             *,
                             allow_truncated : bool,
                             allow_excess: bool) -> str:
    try:
        unpacked_message = message.unpack_container(data,
                                                    allow_truncated=allow_truncated)
        decoded_message = message.decode_container(data,
                                                   decode_choices=decode_choices,
                                                   scaling=True,
                                                   allow_truncated=allow_truncated,
                                                   allow_excess=allow_excess)

    except DecodeError as e:
        return f' {e}'

    if single_line:
        return _format_container_single_line(message,
                                             unpacked_message,
                                             decoded_message)
    else:
        return _format_container_multi_line(message,
                                            unpacked_message,
                                            decoded_message)


def format_message(message : Message,
                   decoded_signals : SignalDictType,
                   single_line : bool) -> str:
    formatted_signals = format_signals(message, decoded_signals)

    if single_line:
        return _format_message_single_line(message.name, formatted_signals)
    else:
        return _format_message_multi_line(message.name, formatted_signals)

def format_multiplexed_name(message : Message,
                            decoded_signals : SignalDictType) -> str:
    # The idea here is that we rely on the sorted order of the Signals, and
    # then simply go through each possible Multiplexer and build a composite
    # key consisting of the Message name prepended to all the possible MUX
    # Signals (and their values). This composite key is therefore unique for
    # all the different possible enumerations of MUX values, which allows us
    # to display each MUXed Message on its own separate line.
    result = [message.name]

    for signal in message.signals:
        if signal.is_multiplexer:
            if signal.name in decoded_signals:
                value = decoded_signals[signal.name]
            elif signal.raw_initial is not None:
                value = signal.raw_initial
            else:
                value = 0
            result.append(f'{signal.name}={value}')

    return ' :: '.join(result)



from typing import Generic, TypeVar

AttributeValue = str | int | float
AttributeValueTypeVar = TypeVar('AttributeValueTypeVar', str, int, float)

class AttributeDefinition(Generic[AttributeValueTypeVar]):
    """A definition of an attribute that can be associated with attributes
    in nodes/messages/signals.

    """

    def __init__(self,
                 name: str,
                 default_value: AttributeValueTypeVar | None = None,
                 kind: str | None = None,
                 type_name: str | None = None,
                 minimum: float | None = None,
                 maximum: float | None = None,
                 choices: list[str] | None = None) -> None:
        self._name = name
        self._default_value: AttributeValueTypeVar | None = default_value
        self._kind = kind
        self._type_name = type_name
        self._minimum = minimum
        self._maximum = maximum
        self._choices = choices or []

    @property
    def name(self) -> str:
        """The attribute name as a string.

        """

        return self._name

    @property
    def default_value(self) -> AttributeValueTypeVar | None:
        """The default value that this attribute has, or ``None`` if
        unavailable.

        """

        return self._default_value

    @default_value.setter
    def default_value(self, value: AttributeValueTypeVar | None) -> None:
        self._default_value = value

    @property
    def kind(self) -> str | None:
        """The attribute kind (BU\\_, BO\\_, SG\\_), or ``None`` if unavailable.

        """

        return self._kind

    @property
    def type_name(self) -> str | None:
        """The attribute type (INT, HEX, FLOAT, STRING, ENUM), or ``None`` if
        unavailable.

        """

        return self._type_name

    @property
    def minimum(self) -> int | float | None:
        """The minimum value of the attribute, or ``None`` if unavailable.

        """

        return self._minimum

    @minimum.setter
    def minimum(self, value: float | None) -> None:
        self._minimum = value

    @property
    def formatted_minimum(self) -> str:
        """The defined minimum value of the attribute definition as it appears in the DBC file

        """
        return '' if self._minimum is None else f' {self._minimum}'

    @property
    def maximum(self) -> int | float | None:
        """The maximum value of the attribute, or ``None`` if unavailable.

        """

        return self._maximum

    @maximum.setter
    def maximum(self, value: float | None) -> None:
        self._maximum = value

    @property
    def formatted_maximum(self) -> str:
        """The defined maximum value of the attribute definition as it appears in the DBC file

        """
        return '' if self._maximum is None else f' {self._maximum}'

    @property
    def choices(self) -> list[str]:
        """A dictionary mapping attribute values to enumerated choices, or
        ``None`` if unavailable.

        """

        return self._choices

    @choices.setter
    def choices(self, value: list[str]) -> None:
        self._choices = value

    def __repr__(self) -> str:
        return f"attribute_definition('{self._name}', {self._default_value})"

AttributeDefinitionType = AttributeDefinition[str] | AttributeDefinition[int] | AttributeDefinition[float]



from typing import Generic

from .attribute_definition import AttributeDefinition, AttributeValueTypeVar


class Attribute(Generic[AttributeValueTypeVar]):
    """An attribute that can be associated with nodes/messages/signals.

    """

    def __init__(self,
                 value: AttributeValueTypeVar,
                 definition: AttributeDefinition[AttributeValueTypeVar]) -> None:
        self._value: AttributeValueTypeVar = value
        self._definition: AttributeDefinition[AttributeValueTypeVar] = definition

    @property
    def name(self) -> str:
        """The attribute name as a string.

        """

        return self._definition.name

    @property
    def value(self) -> AttributeValueTypeVar:
        """The value that this attribute has.

        """

        return self._value

    @value.setter
    def value(self, value: AttributeValueTypeVar) -> None:
        self._value = value

    @property
    def formatted_value(self) -> str:
        """The value of the attribute as it appears in the DBC file

        """
        if self.definition.type_name == "STRING":
            return f'"{self._value}"'

        return f'{self._value}'

    @property
    def definition(self) -> AttributeDefinition[AttributeValueTypeVar]:
        """The attribute definition.

        """

        return self._definition

    def __repr__(self) -> str:
        return f"attribute('{self.name}', {self.value})"


AttributeType = Attribute[str] | Attribute[int] | Attribute[float]



__all__ = ["Bus", "Database", "DecodeError", "EncodeError", "Message",
           "Node", "Signal", "dump_file", "load", "load_file", "load_string"]

import os
import warnings
from contextlib import nullcontext
from typing import Any, TextIO

try:
    import diskcache  # type: ignore
    _DISKCACHE_AVAILABLE = True
except ImportError:
    _DISKCACHE_AVAILABLE = False

from ..typechecking import StringPathLike
from . import can, diagnostics, utils
from .can.bus import Bus
from .can.database import Database
from .can.message import Message
from .can.node import Node
from .can.signal import Signal
from .errors import (
    DecodeError,
    EncodeError,
    Error,
    UnsupportedDatabaseFormatError,
)


def _resolve_database_format_and_encoding(database_format: str | None,
                                          encoding: str | None,
                                          filename: StringPathLike) -> tuple[str, str]:
    if database_format is None:
        database_format = os.path.splitext(filename)[1][1:].lower()

    if encoding is None:
        try:
            encoding = {
                'dbc': 'cp1252',
                'sym': 'cp1252'
            }[database_format]
        except KeyError:
            encoding = 'utf-8'

    return database_format, encoding

def load_file(filename: StringPathLike,
              database_format: str | None = None,
              encoding: str | None = None,
              frame_id_mask: int | None = None,
              prune_choices: bool = False,
              strict: bool = True,
              cache_dir: str | None = None,
              sort_signals: utils.type_sort_signals = utils.sort_signals_by_start_bit,
              ) -> can.Database | diagnostics.Database:
    """Open, read and parse given database file and return a
    :class:`can.Database<.can.Database>` or
    :class:`diagnostics.Database<.diagnostics.Database>` object with
    its contents.

    `database_format` is one of ``'arxml'``, ``'dbc'``, ``'kcd'``,
    ``'sym'``, ``cdd`` and ``None``. If ``None``, the database format
    is selected based on the filename extension as in the table below.
    Filename extensions are case insensitive.

    +-----------+-----------------+
    | Extension | Database format |
    +===========+=================+
    | .arxml    | ``'arxml'``     |
    +-----------+-----------------+
    | .dbc      | ``'dbc'``       |
    +-----------+-----------------+
    | .kcd      | ``'kcd'``       |
    +-----------+-----------------+
    | .sym      | ``'sym'``       |
    +-----------+-----------------+
    | .cdd      | ``'cdd'``       |
    +-----------+-----------------+
    | <unknown> | ``None``        |
    +-----------+-----------------+

    `encoding` specifies the file encoding. If ``None``, the encoding
    is selected based on the database format as in the table
    below. Use ``open()`` and :func:`~cantools.database.load()` if
    platform dependent encoding is desired.

    +-----------------+-------------------+
    | Database format | Default encoding  |
    +=================+===================+
    | ``'arxml'``     | ``'utf-8'``       |
    +-----------------+-------------------+
    | ``'dbc'``       | ``'cp1252'``      |
    +-----------------+-------------------+
    | ``'kcd'``       | ``'utf-8'``       |
    +-----------------+-------------------+
    | ``'sym'``       | ``'cp1252'``      |
    +-----------------+-------------------+
    | ``'cdd'``       | ``'utf-8'``       |
    +-----------------+-------------------+
    | ``None``        | ``'utf-8'``       |
    +-----------------+-------------------+

    `prune_choices` abbreviates the names of choices by removing
    a common prefix ending on an underscore. If you want to have
    the original names you need to pass `prune_choices = False`.

    `cache_dir` specifies the database cache location in the file
    system. Give as ``None`` to disable the cache. By default the
    cache is disabled, but can be enabled with environment variable
    `CANTOOLS_CACHE_DIR`. The cache key is db path with modification
    time and all arguments that may influence the result. Using a
    cache will significantly reduce the load time when reloading the
    same file. The cache directory is automatically created if it does
    not exist. Remove the cache directory `cache_dir` to clear the cache.
    Requires the optional ``diskcache`` package
    (``pip install cantools[cache]``); if it is not installed, a
    warning is issued and caching is disabled.

    See :func:`~cantools.database.load_string()` for descriptions of
    other arguments.

    Raises an
    :class:`~cantools.database.UnsupportedDatabaseFormatError`
    exception if given file does not contain a supported database
    format.

    >>> db = cantools.database.load_file('foo.dbc')
    >>> db.version
    '1.0'

    """

    database_format, encoding = _resolve_database_format_and_encoding(
        database_format,
        encoding,
        filename)

    cache_dir = cache_dir or os.getenv("CANTOOLS_CACHE_DIR", None)
    if cache_dir and not _DISKCACHE_AVAILABLE:
        warnings.warn(
            "diskcache is not installed; caching is disabled. "
            "Install it with: pip install cantools[cache]",
            UserWarning,
            stacklevel=2,
        )
        cache_dir = None
    cache_key: tuple[Any, ...] | None = None
    db: can.Database | diagnostics.Database

    with diskcache.Cache(cache_dir) if cache_dir else nullcontext() as cache:
        if cache is not None:
            # do not cache if user-defined sort_signals function is provided
            # the key cannot be created if function is local or depends on context
            # pickle serializer will fail anyway
            if not callable(sort_signals) or sort_signals.__module__ == 'cantools.database.utils':
                cache_key = (
                    database_format,
                    encoding,
                    frame_id_mask,
                    prune_choices,
                    strict,
                    sort_signals,
                    filename,
                    os.path.getmtime(filename),
                )

            db = cache.get(cache_key)
            if isinstance(db, (can.Database, diagnostics.Database)):
                return db

        with open(filename, encoding=encoding, errors='replace') as fin:
            db = load(fin,
                    database_format,
                    frame_id_mask,
                    prune_choices,
                    strict,
                    sort_signals)

        if cache is not None:
            cache[cache_key] = db

        return db

def dump_file(database: can.Database,
              filename: StringPathLike,
              database_format: str | None = None,
              encoding: str | None = None,
              sort_signals: utils.type_sort_signals = utils.SORT_SIGNALS_DEFAULT) -> None:
    """Dump given database `database` to given file `filename`.

    Depending on the output file format signals may be sorted by default.
    If you don't want signals to be sorted pass `sort_signals=None`.
    `sort_signals=None` is assumed by default if you have passed `sort_signals=None` to load_file.
    If you want the signals to be sorted in a special way pass something like
    `sort_signals = lambda signals: list(sorted(signals, key=lambda sig: sig.name))`
    For dbc files the default is to sort the signals by their start bit in descending order.
    For kcd files the default is to not sort the signals.

    See :func:`~cantools.database.load_file()` for descriptions of
    other arguments.

    The ``'dbc'`` database format will always have Windows-style line
    endings (``\\r\\n``). For other database formats the line ending
    depends on the operating system.

    >>> db = cantools.database.load_file('foo.dbc')
    >>> cantools.database.dump_file(db, 'bar.dbc')

    Pass `sort_signals=None, prune_choices=False` to load_file
    in order to minimize the differences between foo.dbc and bar.dbc.
    """

    database_format, encoding = _resolve_database_format_and_encoding(
        database_format,
        encoding,
        filename)

    newline = None

    if database_format == 'dbc':
        output = database.as_dbc_string(sort_signals=sort_signals)
        newline = ''
    elif database_format == 'kcd':
        output = database.as_kcd_string(sort_signals=sort_signals)
    elif database_format == 'sym':
        output = database.as_sym_string(sort_signals=sort_signals)
    else:
        raise Error(
            f"Unsupported output database format '{database_format}'.")

    with open(filename, 'w', encoding=encoding, newline=newline, errors='replace') as fout:
        fout.write(output)


def load(fp: TextIO,
         database_format: str | None = None,
         frame_id_mask: int | None = None,
         prune_choices: bool = False,
         strict: bool = True,
         sort_signals: utils.type_sort_signals = utils.sort_signals_by_start_bit) -> can.Database | diagnostics.Database:
    """Read and parse given database file-like object and return a
    :class:`can.Database<.can.Database>` or
    :class:`diagnostics.Database<.diagnostics.Database>` object with
    its contents.

    See :func:`~cantools.database.load_string()` for descriptions of
    other arguments.

    Raises an
    :class:`~cantools.database.UnsupportedDatabaseFormatError`
    exception if given file-like object does not contain a supported
    database format.

    >>> with open('foo.kcd') as fin:
    ...    db = cantools.database.load(fin)
    >>> db.version
    None

    """

    return load_string(fp.read(),
                       database_format,
                       frame_id_mask,
                       prune_choices,
                       strict,
                       sort_signals)


def load_string(string: str,
                database_format: str | None = None,
                frame_id_mask: int | None = None,
                prune_choices: bool = False,
                strict: bool = True,
                sort_signals: utils.type_sort_signals = utils.sort_signals_by_start_bit) \
        -> can.Database | diagnostics.Database:
    """Parse given database string and return a
    :class:`can.Database<.can.Database>` or
    :class:`diagnostics.Database<.diagnostics.Database>` object with
    its contents.

    `database_format` may be one of ``'arxml'``, ``'dbc'``, ``'kcd'``,
    ``'sym'``, ``'cdd'`` or ``None``, where ``None`` means transparent
    format.

    `prune_choices` is a bool indicating whether signal names are supposed to be abbreviated
    by stripping a common prefix ending on an underscore. This is enabled by default.

    See :class:`can.Database<.can.Database>` for a description of
    `strict`.

    `sort_signals` is a function taking a list of signals as argument and returning a list of signals.
    By default signals are sorted by their start bit when their Message object is created.
    If you don't want them to be sorted pass `sort_signals = None`.
    If you want the signals to be sorted in another way pass something like
    `sort_signals = lambda signals: list(sorted(signals, key=lambda sig: sig.name))`

    Raises an
    :class:`~cantools.database.UnsupportedDatabaseFormatError`
    exception if given string does not contain a supported database
    format.

    >>> with open('foo.dbc') as fin:
    ...    db = cantools.database.load_string(fin.read())
    >>> db.version
    '1.0'

    """

    if database_format not in ['arxml', 'dbc', 'kcd', 'sym', 'cdd', None]:
        raise ValueError(
            f"expected database format 'arxml', 'dbc', 'kcd', 'sym', 'cdd' or "
            f"None, but got '{database_format}'")

    e_arxml = None
    e_dbc = None
    e_kcd = None
    e_sym = None
    e_cdd = None

    def load_can_database(fmt: str) -> can.Database:
        db = can.Database(frame_id_mask=frame_id_mask,
                          strict=strict,
                          sort_signals=sort_signals)

        if fmt == 'arxml':
            db.add_arxml_string(string)
        elif fmt == 'dbc':
            db.add_dbc_string(string)
        elif fmt == 'kcd':
            db.add_kcd_string(string)
        elif fmt == 'sym':
            db.add_sym_string(string)

        if prune_choices:
            utils.prune_database_choices(db)

        return db

    if database_format in ['arxml', None]:
        try:
            return load_can_database('arxml')
        except Exception as e:
            e_arxml = e

    if database_format in ['dbc', None]:
        try:
            return load_can_database('dbc')
        except Exception as e:
            e_dbc = e

    if database_format in ['kcd', None]:
        try:
            return load_can_database('kcd')
        except Exception as e:
            e_kcd = e

    if database_format in ['sym', None]:
        try:
            return load_can_database('sym')
        except Exception as e:
            e_sym = e

    if database_format in ['cdd', None]:
        try:
            db = diagnostics.Database()
            db.add_cdd_string(string)
            return db
        except Exception as e:
            e_cdd = e

    if database_format is not None:
        # raise an error while keeping the traceback of the original
        # exception usable. note that for this we cannot auto-detect
        # the format because the probing mechanism raises an exception
        # for every single supported database format in this case
        exc = e_arxml or e_dbc or e_kcd or e_sym or e_cdd
        raise UnsupportedDatabaseFormatError(e_arxml,
                                             e_dbc,
                                             e_kcd,
                                             e_sym,
                                             e_cdd) from exc
    else:
        raise UnsupportedDatabaseFormatError(e_arxml, e_dbc, e_kcd, e_sym, e_cdd)



# DID data.

from cantools.database.conversion import BaseConversion, IdentityConversion
from cantools.database.namedsignalvalue import NamedSignalValue
from cantools.typechecking import ByteOrder, Choices, SignalValueType


class Data:
    """A data data with position, size, unit and other information. A data
    is part of a DID.

    """

    def __init__(self,
                 name: str,
                 start: int,
                 length: int,
                 byte_order: ByteOrder = 'little_endian',
                 conversion: BaseConversion | None = None,
                 minimum: float | None = None,
                 maximum: float | None = None,
                 unit: str | None = None,
                 ) -> None:
        #: The data name as a string.
        self.name: str = name

        #: The conversion instance, which is used to convert
        #: between raw and scaled/physical values.
        self.conversion = conversion or IdentityConversion(is_float=False)

        #: The start bit position of the data within its DID.
        self.start: int = start

        #: The length of the data in bits.
        self.length = length

        #: Data byte order as ``'little_endian'`` or ``'big_endian'``.
        self.byte_order: ByteOrder = byte_order

        #: The minimum value of the data, or ``None`` if unavailable.
        self.minimum: float | None = minimum

        #: The maximum value of the data, or ``None`` if unavailable.
        self.maximum: float | None = maximum

        #: The unit of the data as a string, or ``None`` if unavailable.
        self.unit = unit

        # ToDo: Remove once types are handled properly.
        self.is_signed: bool = False

    def raw_to_scaled(
        self, raw_value: float, decode_choices: bool = True
    ) -> SignalValueType:
        """Convert an internal raw value according to the defined scaling or value table.

        :param raw_value:
            The raw value
        :param decode_choices:
            If `decode_choices` is ``False`` scaled values are not
            converted to choice strings (if available).
        :return:
            The calculated scaled value
        """
        return self.conversion.raw_to_scaled(raw_value, decode_choices)

    def scaled_to_raw(self, scaled_value: SignalValueType) -> int | float:
        """Convert a scaled value to the internal raw value.

        :param scaled_value:
            The scaled value.
        :return:
            The internal raw value.
        """
        return self.conversion.scaled_to_raw(scaled_value)

    def choice_to_number(self, string: str | NamedSignalValue) -> int:
        try:
            return self.conversion.choice_to_number(string)
        except KeyError as exc:
            err_msg = f"Choice {string} not found in Data {self.name}."
            raise KeyError(err_msg) from exc

    @property
    def scale(self) -> int | float:
        """The scale factor of the signal value."""
        return self.conversion.scale

    @scale.setter
    def scale(self, value: float) -> None:
        self.conversion = self.conversion.factory(
            scale=value,
            offset=self.conversion.offset,
            choices=self.conversion.choices,
            is_float=self.conversion.is_float,
        )

    @property
    def offset(self) -> int | float:
        """The offset of the signal value."""
        return self.conversion.offset

    @offset.setter
    def offset(self, value: float) -> None:
        self.conversion = self.conversion.factory(
            scale=self.conversion.scale,
            offset=value,
            choices=self.conversion.choices,
            is_float=self.conversion.is_float,
        )

    @property
    def choices(self) -> Choices | None:
        """A dictionary mapping signal values to enumerated choices, or
        ``None`` if unavailable."""
        return self.conversion.choices

    @choices.setter
    def choices(self, choices: Choices | None) -> None:
        self.conversion = self.conversion.factory(
            scale=self.conversion.scale,
            offset=self.conversion.offset,
            choices=choices,
            is_float=self.conversion.is_float,
        )

    @property
    def is_float(self) -> bool:
        """``True`` if the raw signal value is a float, ``False`` otherwise."""
        return self.conversion.is_float

    @is_float.setter
    def is_float(self, is_float: bool) -> None:
        self.conversion = self.conversion.factory(
            scale=self.conversion.scale,
            offset=self.conversion.offset,
            choices=self.conversion.choices,
            is_float=is_float,
        )

    def __repr__(self) -> str:
        if self.choices is None:
            choices = None
        else:
            choices = '{{{}}}'.format(', '.join(
                [f"{value}: '{text}'"
                 for value, text in self.choices.items()]))

        return f"data('{self.name}', {self.start}, {self.length}, '{self.byte_order}', {self.conversion.scale}, {self.conversion.offset}, {self.minimum}, {self.maximum}, '{self.unit}', {choices})"



# Utilities for calculating the CRC of the AUTOSAR end-to-end
# protection specification


import crccheck  # type: ignore

from ..database.can.message import Message


def compute_profile2_crc(payload: bytes,
                         msg_or_data_id: int | Message) -> int | None:
    """Compute the CRC checksum for profile 2 of the AUTOSAR end-to-end
    protection specification.

    data_id is the data ID to be used. If it is unspecified, it is
    determined from the message's ``autosar.e2e.data_ids`` attribute.
    """

    if len(payload) < 2:
        # Profile 2 E2E protection requires at least 2 bytes
        return None

    protected_len = None
    data_id = None

    if isinstance(msg_or_data_id, Message):
        msg = msg_or_data_id
        if msg.autosar is None or \
           msg.autosar.e2e is None or \
           msg.autosar.e2e.data_ids is None or \
           len(msg.autosar.e2e.data_ids) != 16:
            # message is not end-to-end protected using profile 2
            return None

        assert msg.autosar is not None
        assert msg.autosar.e2e is not None
        assert msg.autosar.e2e.data_ids is not None

        protected_len = msg.autosar.e2e.payload_length
        seq_counter = payload[1] & 0xf
        data_id = msg.autosar.e2e.data_ids[seq_counter]
    else:
        protected_len = len(payload)
        data_id = msg_or_data_id

    # create the data to be checksummed
    crc_data = bytearray(payload[1:protected_len])

    # append data id
    crc_data += bytearray([ data_id ])

    # do the actual work
    return int(crccheck.crc.Crc8Autosar().calc(crc_data))

def apply_profile2_crc(payload: bytes,
                       msg_or_data_id: int | Message) \
  -> bytearray | None:
    """Compute the CRC checksum for profile 2 of the AUTOSAR end-to-end
    protection specification and apply it to an encoded payload.

    If the message is passed, this function also takes care of special
    cases like the message not being end-to-end protected or being a
    secured frame.
    """

    crc = compute_profile2_crc(payload, msg_or_data_id)

    if crc is None:
        return None

    result = bytearray(payload)
    result[0] = crc
    return result


def check_profile2_crc(payload: bytes,
                       msg_or_data_id: int | Message) -> bool | None:
    """Check if the AUTOSAR E2E checksum for profile 2 of the AUTOSAR
    end-to-end protection specification is correct.

    If a message is not end-to-end protected by profile 2, ``False`` is
    returned.
    """

    crc = compute_profile2_crc(payload, msg_or_data_id)

    if crc is None:
        return None

    crc2 = payload[0]

    return crc == crc2

def compute_profile5_crc(payload: bytes,
                         msg_or_data_id: int | Message) -> int | None:
    """Compute the CRC checksum for profile 5 of the AUTOSAR end-to-end
    protection specification.

    data_id is the data ID to be used. If it is unspecified, it is
    determined from the message's ``autosar.e2e.data_ids`` attribute.
    """

    if len(payload) < 4:
        # Profile 5 E2E protection requires at least 4 bytes
        return None

    protected_len = None
    data_id = None

    if isinstance(msg_or_data_id, Message):
        msg = msg_or_data_id
        if msg_or_data_id.autosar is None or \
           msg_or_data_id.autosar.e2e is None or \
           msg_or_data_id.autosar.e2e.data_ids is None or \
           len(msg_or_data_id.autosar.e2e.data_ids) != 1:
            # message is not end-to-end protected using profile 5
            return None

        assert msg.autosar is not None
        assert msg.autosar.e2e is not None
        assert msg.autosar.e2e.data_ids is not None

        protected_len = msg.autosar.e2e.payload_length
        data_id = msg.autosar.e2e.data_ids[0]
    else:
        protected_len = len(payload)
        data_id = msg_or_data_id

    # we assume that the "offset" parameter given in the specification
    # is always 0...
    result = crccheck.crc.Crc16Autosar().calc(payload[2:protected_len],
                                              initvalue=0xffff)

    # deal with the data id
    result = crccheck.crc.Crc16Autosar().calc(bytearray([data_id&0xff]),
                                              initvalue=result)
    result = crccheck.crc.Crc16Autosar().calc(bytearray([(data_id>>8)&0xff]),
                                              initvalue=result)

    return int(result)

def apply_profile5_crc(payload: bytes,
                       msg_or_data_id: int | Message) \
  -> bytearray | None:
    """Compute the AUTOSAR E2E checksum for profile 5 of the AUTOSAR
    end-to-end protection specification and apply it to an encoded
    payload.

    If the message is passed, this function also takes care of special
    cases like the message not being end-to-end protected or being a
    secured frame.

    """

    crc = compute_profile5_crc(payload, msg_or_data_id)

    if crc is None:
        return None

    result = bytearray(payload)
    result[0] = crc&0xff
    result[1] = (crc>>8)&0xff

    return result

def check_profile5_crc(payload: bytes,
                       msg_or_data_id: int | Message) -> bool | None:
    """Check if the AUTOSAR E2E checksum for profile 5 of the AUTOSAR
    end-to-end protection specification is correct.

    If a message is not end-to-end protected by profile 5, ``False`` is
    returned.
    """

    crc = compute_profile5_crc(payload, msg_or_data_id)

    if crc is None:
        return None

    crc2 = payload[0] + (payload[1]<<8)

    return crc == crc2



#!/usr/bin/env python3
#
# > python3 hello_world.py
# Message: {'Temperature': 250.1, 'AverageRadius': 3.2, 'Enable': 'Enabled'}
# Encoded: c001400000000000
# Decoded: {'Enable': 'Enabled', 'AverageRadius': 3.2, 'Temperature': 250.1}
#

import os
from binascii import hexlify

import cantools

SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
MOTOHAWK_PATH = os.path.join(SCRIPT_DIR,
                             '..',
                             'tests',
                             'files',
                             'dbc',
                             'motohawk.dbc')

database = cantools.db.load_file(MOTOHAWK_PATH)

message = {
    'Temperature': 250.1,
    'AverageRadius': 3.2,
    'Enable': 'Enabled'
}

encoded = database.encode_message('ExampleMessage', message)
decoded = database.decode_message('ExampleMessage', encoded)

print('Message:', message)
print('Encoded:', hexlify(encoded).decode('ascii'))
print('Decoded:', decoded)



# Internal CAN database.


from .bus import Bus
from .formats.arxml.database_specifics import AutosarDatabaseSpecifics
from .formats.dbc_specifics import DbcSpecifics
from .message import Message
from .node import Node


class InternalDatabase:
    """Internal CAN database.

    """

    def __init__(self,
                 messages: list[Message],
                 nodes: list[Node],
                 buses: list[Bus],
                 version : str | None,
                 dbc_specifics: DbcSpecifics | None = None,
                 autosar_specifics: AutosarDatabaseSpecifics | None = None):
        self.messages = messages
        self.nodes = nodes
        self.buses = buses
        self.version = version
        self.dbc = dbc_specifics
        self.autosar = autosar_specifics
